# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from time import monotonic
from typing import Dict
from collections.abc import Callable

from xpra.util.str_fn import strtobytes
from xpra.gtk.error import XError, xsync
from xpra.x11.common import X11Event
from xpra.util.str_fn import csv

from xpra.log import Logger
log = Logger("x11", "bindings")
verbose = Logger("x11", "bindings", "verbose")


from xpra.x11.bindings.xlib cimport (
    Display, Window, Visual, XID, XRectangle, Atom, Time, CARD32, Bool,
    XEvent, XSelectionRequestEvent, XSelectionClearEvent,
    XSelectionEvent,
    XFree, XGetErrorText,
    XQueryTree,
    XDefaultRootWindow,
    XGetAtomName,
)
from libc.stdint cimport uintptr_t


DEF XNone = 0

cdef extern from "X11/Xlib.h":
    int NotifyNormal

    int BadWindow
    int MapRequest
    int ConfigureRequest
    int SelectionRequest
    int SelectionClear
    int FocusIn
    int FocusOut
    int KeymapNotify
    int Expose
    int GraphicsExpose
    int NoExpose
    int VisibilityNotify
    int ClientMessage
    int CreateNotify
    int MapNotify
    int UnmapNotify
    int DestroyNotify
    int ConfigureNotify
    int ReparentNotify
    int GravityNotify
    int ResizeRequest
    int CirculateNotify
    int CirculateRequest
    int SelectionNotify
    int ColormapNotify
    int MappingNotify
    int PropertyNotify
    int KeyPress
    int KeyRelease
    int ButtonPress
    int ButtonRelease
    int EnterNotify
    int LeaveNotify
    int MotionNotify
    int GenericEvent


cdef extern from "X11/extensions/xfixeswire.h":
    unsigned int XFixesCursorNotify
    unsigned long XFixesDisplayCursorNotifyMask
    unsigned int XFixesSelectionNotify


cdef extern from "X11/extensions/shape.h":
    Bool XShapeQueryExtension(Display *display, int *event_base, int *error_base)
    ctypedef struct XShapeEvent:
        Window window
        int kind            #ShapeBounding or ShapeClip
        int x, y            #extents of new region
        unsigned width, height
        Bool shaped         #true if the region exists


cdef extern from "X11/extensions/Xdamage.h":
    ctypedef XID Damage
    unsigned int XDamageNotify
    ctypedef struct XDamageNotifyEvent:
        Damage damage
        int level
        Bool more
        XRectangle area
    Bool XDamageQueryExtension(Display *, int * event_base, int * error_base)


cdef extern from "X11/extensions/Xfixes.h":
    Bool XFixesQueryExtension(Display *, int *event_base, int *error_base)
    ctypedef struct XFixesCursorImage:
        short x
        short y
        unsigned short width
        unsigned short height
        unsigned short xhot
        unsigned short yhot
        unsigned long cursor_serial
        unsigned long* pixels
        Atom atom
        char* name
    ctypedef struct XFixesCursorNotifyEvent:
        Window window
        int subtype
        unsigned long cursor_serial
        Time timestamp
        Atom cursor_name

    ctypedef struct XFixesSelectionNotifyEvent:
        int subtype
        Window window
        Window owner
        Atom selection
        Time timestamp
        Time selection_timestamp


cdef extern from "X11/extensions/XKB.h":
    unsigned int XkbUseCoreKbd
    unsigned int XkbBellNotifyMask
    unsigned int XkbBellNotify
    ctypedef struct XkbAnyEvent:
        unsigned long   serial
        Bool            send_event
        Display *       display
        Time            time
        int             xkb_type
        unsigned int    device


cdef extern from "X11/XKBlib.h":
    Bool XkbQueryExtension(Display *, int *opcodeReturn, int *eventBaseReturn, int *errorBaseReturn, int *majorRtrn, int *minorRtrn)


cdef extern from "X11/extensions/XKBproto.h":
    ctypedef struct XkbBellNotifyEvent:
        int         type
        CARD32      serial
        Bool        send_event
        Display*    display
        Time        time
        int         xkb_type
        int         device
        int         percent
        int         pitch
        int         duration
        int         bell_class
        int         bell_id
        Atom        name
        Window      window
        Bool        event_only


cdef int CursorNotify = -1
cdef int XKBNotify = -1
cdef int ShapeNotify = -1
cdef int XFSelectionNotify = -1


cdef int get_XKB_event_base(Display *xdisplay) noexcept:
    cdef int opcode = 0
    cdef int event_base = 0
    cdef int error_base = 0
    cdef int major = 0
    cdef int minor = 0
    if not XkbQueryExtension(xdisplay, &opcode, &event_base, &error_base, &major, &minor):
        log.warn("Warning: Xkb extension is not available")
        return -1
    verbose(f"get_XKB_event_base(%#x)=%i", <uintptr_t> xdisplay, event_base)
    return event_base


cdef int get_XFixes_event_base(Display *xdisplay) noexcept:
    cdef int event_base = 0
    cdef int error_base = 0
    if not XFixesQueryExtension(xdisplay, &event_base, &error_base):
        log.warn("Warning: XFixes extension is not available")
        return -1
    verbose("get_XFixes_event_base(%#x)=%i", <uintptr_t> xdisplay, event_base)
    assert event_base>0, "invalid event base for XFixes"
    return event_base


cdef int get_XDamage_event_base(Display *xdisplay) noexcept:
    cdef int event_base = 0
    cdef int error_base = 0
    if not XDamageQueryExtension(xdisplay, &event_base, &error_base):
        log.warn("Warning: XDamage extension is not available")
        return -1
    verbose("get_XDamage_event_base(%#x)=%i", <uintptr_t> xdisplay, event_base)
    assert event_base>0, "invalid event base for XDamage"
    return event_base


cdef int get_XShape_event_base(Display *xdisplay) noexcept:
    cdef int event_base = 0, ignored = 0
    if not XShapeQueryExtension(xdisplay, &event_base, &ignored):
        log.warn("Warning: XShape extension is not available")
        return -1
    return event_base


cdef void init_x11_events(Display *display):
    add_x_event_signals({
        MapRequest          : (None, "x11-child-map-request-event"),
        ConfigureRequest    : (None, "x11-child-configure-request-event"),
        SelectionRequest    : ("x11-selection-request", None),
        SelectionClear      : ("x11-selection-clear", None),
        FocusIn             : ("x11-focus-in-event", None),
        FocusOut            : ("x11-focus-out-event", None),
        ClientMessage       : ("x11-client-message-event", None),
        CreateNotify        : ("x11-create-event", None),
        MapNotify           : ("x11-map-event", "x11-child-map-event"),
        UnmapNotify         : ("x11-unmap-event", "x11-child-unmap-event"),
        DestroyNotify       : ("x11-destroy-event", None),
        ConfigureNotify     : ("x11-configure-event", None),
        ReparentNotify      : ("x11-reparent-event", None),
        PropertyNotify      : ("x11-property-notify-event", None),
        KeyPress            : ("x11-key-press-event", None),
        EnterNotify         : ("x11-enter-event", None),
        LeaveNotify         : ("x11-leave-event", None),
        MotionNotify        : ("x11-motion-event", None)       #currently unused, just defined for debugging purposes
    })
    add_x_event_type_names({
        KeyPress            : "KeyPress",
        KeyRelease          : "KeyRelease",
        ButtonPress         : "ButtonPress",
        ButtonRelease       : "ButtonRelease",
        MotionNotify        : "MotionNotify",
        EnterNotify         : "EnterNotify",
        LeaveNotify         : "LeaveNotify",
        FocusIn             : "FocusIn",
        FocusOut            : "FocusOut",
        KeymapNotify        : "KeymapNotify",
        Expose              : "Expose",
        GraphicsExpose      : "GraphicsExpose",
        NoExpose            : "NoExpose",
        VisibilityNotify    : "VisibilityNotify",
        CreateNotify        : "CreateNotify",
        DestroyNotify       : "DestroyNotify",
        UnmapNotify         : "UnmapNotify",
        MapNotify           : "MapNotify",
        MapRequest          : "MapRequest",
        ReparentNotify      : "ReparentNotify",
        ConfigureNotify     : "ConfigureNotify",
        ConfigureRequest    : "ConfigureRequest",
        GravityNotify       : "GravityNotify",
        ResizeRequest       : "ResizeRequest",
        CirculateNotify     : "CirculateNotify",
        CirculateRequest    : "CirculateRequest",
        PropertyNotify      : "PropertyNotify",
        SelectionClear      : "SelectionClear",
        SelectionRequest    : "SelectionRequest",
        SelectionNotify     : "SelectionNotify",
        ColormapNotify      : "ColormapNotify",
        ClientMessage       : "ClientMessage",
        MappingNotify       : "MappingNotify",
        GenericEvent        : "GenericEvent",
    })
    cdef int event_base = get_XShape_event_base(display)
    if event_base>=0:
        global ShapeNotify
        ShapeNotify = event_base
        add_x_event_signal(ShapeNotify, ("x11-shape-event", None))
        add_x_event_type_name(ShapeNotify, "ShapeNotify")
        log("added ShapeNotify=%s", ShapeNotify)
    event_base = get_XKB_event_base(display)
    if event_base>=0:
        global XKBNotify
        XKBNotify = event_base
        add_x_event_signal(XKBNotify, ("x11-xkb-event", None))
        add_x_event_type_name(XKBNotify, "XKBNotify")
    event_base = get_XFixes_event_base(display)
    if event_base>=0:
        global CursorNotify
        CursorNotify = XFixesCursorNotify+event_base
        add_x_event_signal(CursorNotify, ("x11-cursor-event", None))
        add_x_event_type_name(CursorNotify, "CursorNotify")
        global XFSelectionNotify
        XFSelectionNotify = XFixesSelectionNotify+event_base
        add_x_event_signal(XFSelectionNotify, ("x11-xfixes-selection-notify-event", None))
        add_x_event_type_name(XFSelectionNotify, "XFSelectionNotify")
    event_base = get_XDamage_event_base(display)
    if event_base>0:
        global DamageNotify
        DamageNotify = XDamageNotify+event_base
        add_x_event_signal(DamageNotify, ("x11-damage-event", None))
        add_x_event_type_name(DamageNotify, "DamageNotify")
    set_debug_events()


x_event_signals : Dict[int,tuple] = {}


def add_x_event_signal(event: int, mapping: tuple) -> None:
    x_event_signals[event] = mapping


def add_x_event_signals(event_signals: Dict[int,tuple]) -> None:
    x_event_signals.update(event_signals)


def get_x_event_signals(event: int) -> tuple:
    return x_event_signals.get(event)


x_event_type_names : Dict[int,str] = {}
names_to_event_type : Dict[str,int] = {}


def add_x_event_type_name(event: int, name: str) -> None:
    x_event_type_names[event] = name
    names_to_event_type[name] = event


def add_x_event_type_names(event_type_names: Dict[int, str]) -> None:
    x_event_type_names.update(event_type_names)
    for k,v in event_type_names.items():
        names_to_event_type[v] = k
    verbose("x_event_signals=%s", x_event_signals)
    verbose("event_type_names=%s", x_event_type_names)
    verbose("names_to_event_type=%s", names_to_event_type)


def get_x_event_type_name(event: int) -> str:
    return x_event_type_names.get(event, "")


def set_debug_events() -> None:
    global debug_route_events
    XPRA_X11_DEBUG_EVENTS = os.environ.get("XPRA_X11_DEBUG_EVENTS", "")
    debug_set = set()
    ignore_set = set()
    for n in XPRA_X11_DEBUG_EVENTS.split(","):
        name = n.strip()
        if len(name)==0:
            continue
        if name[0]=="-":
            event_set = ignore_set
            name = name[1:]
        else:
            event_set = debug_set
        if name in ("*", "all"):
            events = names_to_event_type.keys()
        elif name in names_to_event_type:
            events = [name]
        else:
            log("unknown X11 debug event type: %s", name)
            continue
        #add to correct set:
        for e in events:
            event_set.add(e)
    events = debug_set.difference(ignore_set)
    debug_route_events = [names_to_event_type.get(x) for x in events]
    if len(events)>0:
        log.warn("debugging of X11 events enabled for:")
        log.warn(" %s", csv(events))
        log.warn(" event codes: %s", csv(debug_route_events))


x_event_parsers : Dict[int, Callable] = {}


def add_x_event_parser(extension_opcode: int, parser : Callable) -> None:
    x_event_parsers[extension_opcode] = parser


cdef str atom_str(Display *display, Atom atom):
    if not atom:
        return ""
    cdef char* atom_name = NULL
    try:
        with xsync:
            atom_name = XGetAtomName(display, atom)
    except XError:
        log.error(f"Error: invalid atom {atom:x}")
        return ""
    r = ""
    if atom_name!=NULL:
        r = atom_name.decode("latin1")
        XFree(atom_name)
    return r


cdef object parse_xevent(Display *d, XEvent *e):
    cdef XDamageNotifyEvent * damage_e
    cdef XFixesCursorNotifyEvent * cursor_e
    cdef XkbAnyEvent * xkb_e
    cdef XkbBellNotifyEvent * bell_e
    cdef XShapeEvent * shape_e
    cdef XSelectionRequestEvent * selectionrequest_e
    cdef XSelectionClearEvent * selectionclear_e
    cdef XSelectionEvent * selection_e
    cdef XFixesSelectionNotifyEvent * selectionnotify_e

    cdef int etype = e.type
    global x_event_type_names, x_event_signals
    event_type = x_event_type_names.get(etype, etype)
    if e.xany.send_event and etype not in (ClientMessage, UnmapNotify):
        log("parse_xevent ignoring %s send_event", event_type)
        return None

    if etype == GenericEvent:
        global x_event_parsers
        parser = x_event_parsers.get(e.xcookie.extension)
        if parser:
            log("calling GenericEvent parser %s(%s)", parser, <uintptr_t> &e.xcookie)
            return parser(<uintptr_t> &e.xcookie)
        log("no GenericEvent parser for extension %s", e.xcookie.extension)
        return None

    cdef object event_args = x_event_signals.get(etype)
    if event_args is None:
        log("no signal handler for %s", event_type)
        return None
    log("parse_xevent event=%s/%s window=%#x", event_args, event_type, e.xany.window)

    def atom(Atom v) -> str:
        return atom_str(d, v)

    cdef object pyev = X11Event(event_type)
    pyev.type = etype
    pyev.send_event = bool(e.xany.send_event)
    pyev.serial = e.xany.serial
    if etype != XKBNotify:
        pyev.delivered_to = e.xany.window
    if etype == DamageNotify:
        damage_e = <XDamageNotifyEvent*>e
        pyev.window = e.xany.window
        pyev.damage = damage_e.damage
        pyev.x = damage_e.area.x
        pyev.y = damage_e.area.y
        pyev.width = damage_e.area.width
        pyev.height = damage_e.area.height
    elif etype == MapRequest:
        pyev.window = e.xmaprequest.window
    elif etype == ConfigureRequest:
        pyev.window = e.xconfigurerequest.window
        pyev.x = e.xconfigurerequest.x
        pyev.y = e.xconfigurerequest.y
        pyev.width = e.xconfigurerequest.width
        pyev.height = e.xconfigurerequest.height
        pyev.border_width = e.xconfigurerequest.border_width
        # In principle there are two cases here: .above is
        # XNone (i.e. not specified in the original request),
        # or .above is an invalid window (i.e. it was
        # specified by the client, but it specified something
        # weird).  I don't see any reason to handle these
        # differently, though.
        pyev.above = e.xconfigurerequest.above
        pyev.detail = e.xconfigurerequest.detail
        pyev.value_mask = e.xconfigurerequest.value_mask
    elif etype == SelectionRequest:
        selectionrequest_e = <XSelectionRequestEvent*> e
        pyev.window = selectionrequest_e.owner
        pyev.requestor = selectionrequest_e.requestor
        pyev.selection = atom(selectionrequest_e.selection)
        pyev.target = atom(selectionrequest_e.target)
        pyev.property = atom(selectionrequest_e.property)
        pyev.time = int(selectionrequest_e.time)
    elif etype == SelectionClear:
        selectionclear_e = <XSelectionClearEvent*> e
        pyev.window = selectionclear_e.window
        pyev.selection = atom(selectionclear_e.selection)
        pyev.time = int(selectionclear_e.time)
    elif etype == SelectionNotify:
        selection_e = <XSelectionEvent*> e
        pyev.requestor = selection_e.requestor
        pyev.selection = atom(selection_e.selection)
        pyev.target = atom(selection_e.target)
        pyev.property = atom(selection_e.property)
        pyev.time = int(selection_e.time)
    elif etype == XFSelectionNotify:
        selectionnotify_e = <XFixesSelectionNotifyEvent*> e
        pyev.window = selectionnotify_e.window
        pyev.subtype = selectionnotify_e.subtype
        pyev.owner = selectionnotify_e.owner
        pyev.selection = atom(selectionnotify_e.selection)
        pyev.timestamp = int(selectionnotify_e.timestamp)
        pyev.selection_timestamp = int(selectionnotify_e.selection_timestamp)
    elif etype == ResizeRequest:
        pyev.window = e.xresizerequest.window
        pyev.width = e.xresizerequest.width
        pyev.height = e.xresizerequest.height
    elif etype in (FocusIn, FocusOut):
        pyev.window = e.xfocus.window
        pyev.mode = e.xfocus.mode
        pyev.detail = e.xfocus.detail
    elif etype in (EnterNotify, LeaveNotify):
        pyev.window = e.xcrossing.window
        pyev.mode = e.xcrossing.mode
        pyev.detail = e.xcrossing.detail
        pyev.subwindow = e.xcrossing.subwindow
        pyev.focus = bool(e.xcrossing.focus)
    elif etype == ClientMessage:
        pyev.window = e.xany.window
        if int(e.xclient.message_type) > 2**32:
            log.warn("Warning: Xlib claims that this ClientEvent's 32-bit")
            log.warn(f" message_type is {e.xclient.message_type}.")
            log.warn(" note that this is >2^32.")
            log.warn(" this makes no sense, so I'm ignoring it")
            return None
        pyev.message_type = atom(e.xclient.message_type)
        pyev.format = e.xclient.format
        pieces = []
        if pyev.format == 32:
            for i in range(5):
                # Mask with 0xffffffff to prevent sign-extension on
                # architectures where Python's int is 64-bits.
                pieces.append(int(e.xclient.data.l[i]) & 0xffffffff)
        elif pyev.format == 16:
            for i in range(10):
                pieces.append(int(e.xclient.data.s[i]))
        elif pyev.format == 8:
            for i in range(20):
                pieces.append(int(e.xclient.data.b[i]))
        else:
            log.warn(f"Warning: ignoring ClientMessage {pyev.message_type!r} with format={pyev.format}")
            return None
        pyev.data = tuple(pieces)
    elif etype == CreateNotify:
        pyev.window = e.xcreatewindow.window
        pyev.width = e.xcreatewindow.width
        pyev.height = e.xcreatewindow.height
    elif etype == MapNotify:
        pyev.window = e.xmap.window
        pyev.override_redirect = bool(e.xmap.override_redirect)
    elif etype == UnmapNotify:
        pyev.window = e.xunmap.window
        pyev.from_configure = bool(e.xunmap.from_configure)
    elif etype == DestroyNotify:
        pyev.window = e.xdestroywindow.window
    elif etype == PropertyNotify:
        pyev.window = e.xany.window
        pyev.atom = atom(e.xproperty.atom)
        pyev.time = e.xproperty.time
    elif etype == ConfigureNotify:
        pyev.window = e.xconfigure.window
        pyev.x = e.xconfigure.x
        pyev.y = e.xconfigure.y
        pyev.width = e.xconfigure.width
        pyev.height = e.xconfigure.height
        pyev.border_width = e.xconfigure.border_width
        pyev.above = e.xconfigure.above
        pyev.override_redirect = bool(e.xconfigure.override_redirect)
    elif etype == CirculateNotify:
        pyev.window = e.xcirculaterequest.window
        pyev.place = e.xcirculaterequest.place
    elif etype == ReparentNotify:
        pyev.window = e.xreparent.window
    elif etype == KeyPress:
        pyev.window = e.xany.window
        pyev.hardware_keycode = e.xkey.keycode
        pyev.state = e.xkey.state
    elif etype == CursorNotify:
        pyev.window = e.xany.window
        cursor_e = <XFixesCursorNotifyEvent*>e
        pyev.cursor_serial = cursor_e.cursor_serial
        pyev.cursor_name = atom(cursor_e.cursor_name)
    elif etype == MotionNotify:
        pyev.window = e.xmotion.window
        pyev.root = e.xmotion.root
        pyev.subwindow = e.xmotion.subwindow
        pyev.time = e.xmotion.time
        pyev.x = e.xmotion.x
        pyev.y = e.xmotion.y
        pyev.x_root = e.xmotion.x_root
        pyev.y_root = e.xmotion.y_root
        pyev.state = e.xmotion.state
        pyev.is_hint = e.xmotion.is_hint != NotifyNormal
        #pyev.same_screen = bool(e.xmotion.same_screen)
    elif etype == ShapeNotify:
        shape_e = <XShapeEvent*> e
        pyev.window = shape_e.window
        pyev.kind = shape_e.kind
        pyev.x = shape_e.x
        pyev.y = shape_e.y
        pyev.width = shape_e.width
        pyev.height = shape_e.height
        pyev.shaped = bool(shape_e.shaped)
    elif etype == XKBNotify:
        # note we could just cast directly to XkbBellNotifyEvent
        # but this would be dirty, and we may want to catch
        # other types of XKB events in the future
        xkb_e = <XkbAnyEvent*>e
        verbose("XKBNotify event received xkb_type=%s", xkb_e.xkb_type)
        if xkb_e.xkb_type!=XkbBellNotify:
            return None
        bell_e = <XkbBellNotifyEvent*>e
        pyev.subtype = "bell"
        pyev.device = int(bell_e.device)
        pyev.percent = int(bell_e.percent)
        pyev.pitch = int(bell_e.pitch)
        pyev.duration = int(bell_e.duration)
        pyev.bell_class = int(bell_e.bell_class)
        pyev.bell_id = int(bell_e.bell_id)
        # no idea why window is not set in XkbBellNotifyEvent
        # since we can fire it from a specific window
        # but we need one for the dispatch logic, so use root if unset
        if bell_e.window!=0:
            verbose("using bell_e.window=%#x", bell_e.window)
            pyev.window = bell_e.window
        else:
            pyev.window = XDefaultRootWindow(d)
            verbose("bell using root window=%#x", pyev.window)
        pyev.event_only = bool(bell_e.event_only)
        pyev.delivered_to = pyev.window
        pyev.window_model = None
        pyev.bell_name = atom(bell_e.name)
    else:
        log.info("not handled: %s", x_event_type_names.get(etype, etype))
        return None
    return pyev



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
from time import monotonic
from typing import Any

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.common import ConnectionMessage, FULL_INFO
from xpra.os_util import POSIX
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("network")
pinglog = Logger("network", "ping")
bandwidthlog = Logger("bandwidth")

PING_DETAILS = envbool("XPRA_PING_DETAILS", FULL_INFO > 0)
PING_TIMEOUT = envint("XPRA_PING_TIMEOUT", 60)
AUTO_BANDWIDTH_PCT = envint("XPRA_AUTO_BANDWIDTH_PCT", 80)
assert 1 < AUTO_BANDWIDTH_PCT <= 100, "invalid value for XPRA_AUTO_BANDWIDTH_PCT: %i" % AUTO_BANDWIDTH_PCT
BANDWIDTH_DETECTION = envbool("XPRA_BANDWIDTH_DETECTION", True)
MIN_BANDWIDTH = envint("XPRA_MIN_BANDWIDTH", 5 * 1024 * 1024)


def get_socket_bandwidth_limit(protocol) -> int:
    if not protocol:
        return 0
    # auto-detect:
    pinfo = protocol.get_info()
    socket_speed = pinfo.get("socket", {}).get("device", {}).get("speed")
    if not socket_speed:
        return 0
    bandwidthlog("get_socket_bandwidth_limit() socket_speed=%s", socket_speed)
    # auto: use 80% of socket speed if we have it:
    return socket_speed * AUTO_BANDWIDTH_PCT // 100 or 0


class NetworkStateMixin(StubSourceMixin):

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return any((
            caps.boolget("network-state"),
            typedict(caps.dictget("network") or {}).intget("pings") > 0,
            caps.boolget("ping-echo-sourceid"),  # legacy clients
        ))

    def init_state(self) -> None:
        self.last_ping_echoed_time = 0
        self.check_ping_echo_timers: dict[int, int] = {}
        self.jitter = 0
        self.adapter_type = ""
        self.client_load = (0, 0, 0)
        self.client_connection_data: dict[str, Any] = {}
        self.soft_bandwidth_limit = self.bandwidth_limit = self.server_bandwidth_limit
        self.bandwidth_warnings = True
        self.bandwidth_warning_time = 0

    def init_from(self, _protocol, server) -> None:
        self.server_bandwidth_limit = server.bandwidth_limit
        self.bandwidth_detection = server.bandwidth_detection

    def cleanup(self) -> None:
        self.cancel_ping_echo_timers()

    def get_caps(self) -> dict[str, Any]:
        # legacy flag
        return {"ping-echo-sourceid": True}

    def parse_client_caps(self, c: typedict) -> None:
        self.client_connection_data = c.dictget("connection-data", {})
        ccd = typedict(self.client_connection_data)
        self.adapter_type = ccd.strget("adapter-type")
        self.jitter = ccd.intget("jitter", 0)
        bandwidth_limit = c.intget("bandwidth-limit", 0)
        if getattr(self, "mmap_size", 0) > 0:
            log("mmap enabled, ignoring bandwidth-limit")
            self.bandwidth_limit = 0
            self.bandwidth_detection = False
            self.jitter = 0
        else:
            limit = self.server_bandwidth_limit or get_socket_bandwidth_limit(self.protocol)
            self.bandwidth_limit = min(limit, bandwidth_limit)
            if self.bandwidth_detection:
                self.bandwidth_detection = c.boolget("bandwidth-detection", False)
        bandwidthlog("server bandwidth-limit=%s, client bandwidth-limit=%s, value=%s, detection=%s",
                     self.server_bandwidth_limit, bandwidth_limit, self.bandwidth_limit, self.bandwidth_detection)

    def get_info(self) -> dict[str, Any]:
        lpe = 0
        if self.last_ping_echoed_time > 0:
            lpe = int(monotonic() * 1000 - self.last_ping_echoed_time)
        info = {
            "bandwidth-limit": {
                "setting": self.bandwidth_limit or 0,
                "detection": self.bandwidth_detection,
                "actual": self.soft_bandwidth_limit or 0,
            },
            "jitter": self.jitter,
            "last-ping-echo": lpe,
            "adapter-type": self.adapter_type,
        }
        return info

    ######################################################################
    # pings:
    def ping(self) -> None:
        # NOTE: all ping time/echo time/load avg values are in milliseconds
        now_ms = int(1000 * monotonic())
        pinglog("sending ping to %s with time=%s", self.protocol, now_ms)
        self.send_async("ping", now_ms, int(time.time() * 1000), will_have_more=False)
        timeout = PING_TIMEOUT
        self.check_ping_echo_timers[now_ms] = GLib.timeout_add(timeout * 1000,
                                                               self.check_ping_echo_timeout, now_ms, timeout)

    def check_ping_echo_timeout(self, now_ms: int, timeout: int) -> None:
        self.check_ping_echo_timers.pop(now_ms, None)
        if self.is_closed():
            return
        expired = self.last_ping_echoed_time < now_ms
        message = f"waited {timeout} seconds without a response" if expired else ""
        pinglog(f"check_ping_echo_timeout {message}")
        if expired:
            self.disconnect(ConnectionMessage.CLIENT_PING_TIMEOUT, message)

    def cancel_ping_echo_timers(self) -> None:
        timers = self.check_ping_echo_timers.values()
        self.check_ping_echo_timers = {}
        for t in timers:
            GLib.source_remove(t)

    def process_ping(self, time_to_echo, sid) -> None:
        l1, l2, l3 = 0, 0, 0
        cl = -1
        if PING_DETAILS:
            # send back the load average:
            if POSIX:
                fl1, fl2, fl3 = os.getloadavg()
                l1, l2, l3 = int(fl1 * 1000), int(fl2 * 1000), int(fl3 * 1000)
            # and the last client ping latency we measured (if any):
            stats = getattr(self, "statistics", None)
            if stats and stats.client_ping_latency:
                _, cl = stats.client_ping_latency[-1]
                cl = int(1000.0 * cl)
        self.send_async("ping_echo", time_to_echo, l1, l2, l3, cl, sid, will_have_more=False)
        pinglog(f"ping: sending echo for time={time_to_echo} and {sid=}")

    def process_ping_echo(self, packet) -> None:
        echoedtime, l1, l2, l3, server_ping_latency = packet[1:6]
        timer = self.check_ping_echo_timers.pop(echoedtime, None)
        if timer:
            GLib.source_remove(timer)
        self.last_ping_echoed_time = echoedtime
        client_ping_latency = monotonic() - echoedtime / 1000.0
        stats = getattr(self, "statistics", None)
        if stats and 0 < client_ping_latency < 60:
            stats.client_ping_latency.append((monotonic(), client_ping_latency))
        self.client_load = l1, l2, l3
        if 0 <= server_ping_latency < 60000 and stats:
            stats.server_ping_latency.append((monotonic(), server_ping_latency / 1000.0))
        pinglog(f"ping echo client load={self.client_load}, "
                f"latency measured from server={client_ping_latency}ms, from client={server_ping_latency}ms")

    def update_connection_data(self, data) -> None:
        log("update_connection_data(%s)", data)
        if not isinstance(data, dict):
            raise TypeError("connection-data must be a dictionary")
        self.client_connection_data = data

    def update_bandwidth_limits(self) -> None:
        if not self.bandwidth_detection:
            return
        mmap_size = getattr(self, "mmap_size", 0)
        if mmap_size > 0:
            return
        # calculate soft bandwidth limit based on send congestion data:
        bandwidth_limit = 0
        if BANDWIDTH_DETECTION:
            bandwidth_limit = self.statistics.avg_congestion_send_speed
            bandwidthlog("avg_congestion_send_speed=%s", bandwidth_limit)
            if bandwidth_limit > 20 * 1024 * 1024:
                # ignore congestion speed if greater 20Mbps
                bandwidth_limit = 0
        if (self.bandwidth_limit or 0) > 0:
            # command line options could overrule what we detect?
            bandwidth_limit = min(self.bandwidth_limit, bandwidth_limit)
        if bandwidth_limit > 0:
            bandwidth_limit = max(MIN_BANDWIDTH, bandwidth_limit)
        self.soft_bandwidth_limit = bandwidth_limit
        bandwidthlog("update_bandwidth_limits() bandwidth_limit=%s, soft bandwidth limit=%s",
                     self.bandwidth_limit, bandwidth_limit)
        # figure out how to distribute the bandwidth amongst the windows,
        # we use the window size,
        # (we should use the number of bytes actually sent: framerate, compression, etc..)
        window_weight = {}
        for wid, ws in self.window_sources.items():
            weight = 0
            if not ws.suspended:
                ww, wh = ws.window_dimensions
                # try to reserve bandwidth for at least one screen update,
                # and add the number of pixels damaged:
                weight = ww * wh + ws.statistics.get_damage_pixels()
            window_weight[wid] = weight
        bandwidthlog("update_bandwidth_limits() window weights=%s", window_weight)
        total_weight = max(1, sum(window_weight.values()))
        for wid, ws in self.window_sources.items():
            if bandwidth_limit == 0:
                ws.bandwidth_limit = 0
            else:
                weight = window_weight.get(wid, 0)
                ws.bandwidth_limit = max(MIN_BANDWIDTH // 10, bandwidth_limit * weight // total_weight)



# This file is part of Xpra.
# Copyright (C) 2013 Arthur Huillet
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True, always_allow_keywords=False

import os
import sys
import time
from typing import Any, Tuple, List, Dict
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.codecs.constants import CSCSpec, get_subsampling_divs
from xpra.codecs.image import ImageWrapper

from xpra.log import Logger
log = Logger("csc", "cython")

from libc.stdint cimport uint8_t, uintptr_t # pylint: disable=syntax-error
from xpra.buffers.membuf cimport memalign, buffer_context


cdef extern from "Python.h":
    object PyMemoryView_FromMemory(char *mem, Py_ssize_t size, int flags)
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS
    int PyBUF_WRITE

cdef extern from "stdlib.h":
    void free(void *ptr)

cdef inline int roundup(int n, int m) noexcept:
    return (n + m - 1) & ~(m - 1)

#precalculate indexes in native endianness:
cdef uint8_t BGRX_R, BGRX_G, BGRX_B, BGRX_X
cdef uint8_t RGBX_R, RGBX_G, RGBX_B, RGBX_X
cdef uint8_t RGB_R, RGB_G, RGB_B
cdef uint8_t BGR_R, BGR_G, BGR_B

if sys.byteorder=="little":
    BGRX_B, BGRX_G, BGRX_R, BGRX_X = (0, 1, 2, 3)
    RGBX_R, RGBX_G, RGBX_B, RGBX_X = (0, 1, 2, 3)
    BGR_R, BGR_G, BGR_B = (2, 1, 0)
    RGB_R, RGB_G, RGB_B = (0, 1, 2)
else:
    BGRX_B, BGRX_G, BGRX_R, BGRX_X = (0, 1, 2, 3)
    RGBX_R, RGBX_G, RGBX_B, RGBX_X = (0, 1, 2, 3)
    BGR_R, BGR_G, BGR_B = (0, 1, 2)
    RGB_R, RGB_G, RGB_B = (2, 1, 0)

log("csc_cython: %s endian:", sys.byteorder)
log("csc_cython: byteorder(BGRX)=%s", (BGRX_B, BGRX_G, BGRX_R, BGRX_X))
log("csc_cython: byteorder(RGBX)=%s", (RGBX_R, RGBX_G, RGBX_B, RGBX_X))
log("csc_cython: byteorder(RGB)=%s", (RGB_R, RGB_G, RGB_B))
log("csc_cython: byteorder(BGR)=%s", (BGR_R, BGR_G, BGR_B))

#COLORSPACES = {"BGRX" : ["YUV420P"], "YUV420P" : ["RGB", "BGR", "RGBX", "BGRX"], "GBRP" : ["RGBX", "BGRX"] }


def get_CS(in_cs, valid_options) -> Sequence[str]:
    v = os.environ.get("XPRA_CSC_CYTHON_%s_COLORSPACES" % in_cs)
    if v is None:
        return valid_options
    if not v:
        return ()
    env_override = []
    for cs in v.split(","):
        if cs in valid_options:
            env_override.append(cs)
        else:
            log.warn("invalid colorspace override for %s: %s (only supports: %s)", in_cs, cs, valid_options)
    log("environment override for %s: %s", in_cs, env_override)
    return env_override


COLORSPACES: Dict[str, Sequence[str]] = {
    "BGRX"       : get_CS("BGRX",    ["YUV420P", "YUV444P"]),
    "RGBX"       : get_CS("RGBX",    ["YUV420P", "YUV444P"]),
    "BGR"        : get_CS("BGR",     ["YUV420P", "YUV444P"]),
    "RGB"        : get_CS("RGB",     ["YUV420P", "YUV444P"]),
    "YUV420P"    : get_CS("YUV420P", ["RGB", "BGR", "RGBX", "BGRX"]),
    "GBRP"       : get_CS("GBRP",    ["RGBX", "BGRX"]),
    "r210"       : get_CS("r210",    ["YUV420P", "BGR48", "YUV444P10"]),
    "YUV444P10"  : get_CS("YUV444P10", ["r210"]),
    "YUV444P"    : get_CS("YUV444P",  ["BGRX", "RGBX"]),
    "GBRP10"     : get_CS("GBRP10",  ["r210", ]),
}


def get_type() -> str:
    return "cython"


def get_version() -> Tuple[int, int]:
    return (4, 2)


def get_info() -> Dict[str, Any]:
    return {
        "version"   : (4, 1),
    }


def get_specs() -> Sequence[CSCSpec]:
    specs: Sequence[CSCSpec] = []
    for in_cs, out_css in COLORSPACES.items():
        can_scale = True
        width_mask = height_mask = 0xFFFF
        for out_cs in out_css:
            if in_cs in ("r210", "YUV444P") or out_cs=="r210":
                can_scale = False
            elif in_cs == "GBRP10":
                can_scale = False
            if in_cs == "YUV420P":
                #safer not to try to handle odd dimensions as input:
                width_mask = height_mask = 0xFFFE
            # low score as this should be used as fallback only:
            specs.append(CSCSpec(
                    input_colorspace=in_cs, output_colorspaces=(out_cs, ),
                    codec_class=Converter, codec_type=get_type(),
                    quality=50, speed=0, setup_cost=0, min_w=2, min_h=2,
                    max_w=16*1024, max_h=16*1024,
                    can_scale=can_scale,
                    width_mask=width_mask, height_mask=height_mask)
                )
    return specs


class CythonImageWrapper(ImageWrapper):

    def free(self) -> None:
        log("CythonImageWrapper.free() cython_buffer=%#x", <uintptr_t> self.cython_buffer)
        super().free()
        cb = self.cython_buffer
        if cb>0:
            self.cython_buffer = 0
            free(<void *> (<uintptr_t> cb))

    def _cn(self):
        return "CythonImageWrapper"



DEF STRIDE_ROUNDUP = 2

#Pre-calculate some coefficients and define them as constants
#We use integer calculations so everything is multiplied by 2**16
#To get the result as a byte, we just bitshift:

# RGB to YUV full range
# Y = 0.299 * R + 0.587 * G + 0.114 * B
DEF YR = 19595      # 0.299 * 2**16
DEF YG = 38470      # 0.587 * 2**16
DEF YB = 7471       # 0.114 * 2**16
DEF Yc = 0
DEF YC = 0
# U = -0.169 * R - 0.331 * G + 0.500 * B
DEF UR = -11076     #-0.169 * 2**16
DEF UG = -21692     #-0.331 * 2**16
DEF UB = 32768      # 0.500 * 2**16
DEF Uc = 128
DEF UC = 8388608
# V = 0.500 * R - 0.419 * G - 0.081 * B
DEF VR = 32768      # 0.500  * 2**16
DEF VG = -27460     #-0.419 * 2**16
DEF VB = -5308      #-0.081  * 2**16
DEF Vc = 128
DEF VC = 8388608


DEF MAX_CLAMP = 16777216    #2**(16+8)
DEF MAX_CLAMP10 = 67108864  #2**(16+10)


# YUV to RGB full range:
# Y, Cb and Cr are adjusted as:
# Y'  = Y - 0
# Cb' = Cb - 128
# Cr' = Cr - 128
# (see YC, UC and VC above)
# RGB:
# R = 1*Y'                 + 1.403 * Cr'
# G = 1*Y' - 0.344   * Cb' - 0.714 * Cr'
# B = 1*Y' + 1.770   * Cb'
DEF RY = 65536      #1        * 2**16
DEF RU = 0
DEF RV = 91947      #1.403    * 2**16

DEF GY = 65536      #1        * 2**16
DEF GU = -22544     #-0.344   * 2**16
DEF GV = -46793     #-0.714   * 2**16

DEF BY = 65536      #1        * 2**16
DEF BU = 115999     #1.77     * 2**16
DEF BV = 0


# YUV to RGB studio range:
# Y, Cb and Cr are adjusted as:
# Y'  = Y - 16
# Cb' = Cb - 16 - (240 - 16) // 2 = Cb - 128
# Cr' = Cr - 16 - (240 - 16) // 2 = Cr - 128
# (see YC, UC and VC above)
DEF SYc = 16
DEF SUc = 128
DEF SVc = 128
# RGB:
# we need to scale up:
# Y by 255 // (235-16) = 1.1643835616438356
# UV by 255 // (240-16) = 1.1383928571428572
# R = 1.1643835616438356 * Y'                                  + 1.1383928571428572 * 1.403 * Cr'
# G = 1.1643835616438356 * Y' - 1.1383928571428572*0.344 * Cb' - 1.1383928571428572 * 0.714 * Cr'
# B = 1.1643835616438356 * Y' + 1.1383928571428572*1.770 * Cb'
DEF SRY = 76309      #1 * 1.1643835616438356          * 2**16
DEF SRU = 0
DEF SRV = 104672     #1.403 * 1.1383928571428572      * 2**16

DEF SGY = 76309      #1 * 1.1643835616438356          * 2**16
DEF SGU = -25664     #-0.344 * 1.1383928571428572     * 2**16
DEF SGV = -53268     #-0.714 * 1.1383928571428572     * 2**16

DEF SBY = 76309      #1 * 1.1643835616438356          * 2**16
DEF SBU = 132052     #1.77 * 1.1383928571428572       * 2**16
DEF SBV = 0


cdef inline unsigned char clamp(const long value) noexcept nogil:
    if value <= 0:
        return 0
    cdef long v = value + 2**15
    if v >= MAX_CLAMP:
        return 0xff         #2**8-1
    return <unsigned char> (v>>16)


cdef inline unsigned short clamp10(const long value) noexcept nogil:
    if value <= 0:
        return 0
    cdef long v = value + 2**15
    if v >= MAX_CLAMP10:
        return 0x3ff        #2**10-1
    return <unsigned short> (v>>16)


cdef inline unsigned char clamp_studio_UV(const long value) noexcept nogil:
    if value <= 0:
        return 16
    cdef long v = value + 2**15
    if v >= MAX_CLAMP:
        return 240
    # 240-16
    return <unsigned char> (16 + (v // 74898))


cdef inline unsigned char clamp_studio_Y(const long value) noexcept nogil:
    if value <= 0:
        return 16
    cdef long v = value + 2**15
    if v >= MAX_CLAMP:
        return 235
    # 235-16
    return <unsigned char> (16 + (v // 76608))


cdef inline void r210_to_BGR48_copy(unsigned short *bgr48, const unsigned int *r210,
                                    unsigned int w, unsigned int h,
                                    unsigned int src_stride, unsigned int dst_stride) noexcept nogil:
    cdef unsigned int y
    cdef unsigned int i
    cdef unsigned int v
    for y in range(h):
        i = y*dst_stride//2
        for x in range(w):
            v = r210[x]
            bgr48[i] = v&0x000003ff
            bgr48[i+1] = (v&0x000ffc00) >> 10
            bgr48[i+2] = (v&0x3ff00000) >> 20
            i += 3
        r210 = <unsigned int*> ((<uintptr_t> r210) + src_stride)


cdef inline void gbrp10_to_r210_copy(uintptr_t r210, uintptr_t[3] gbrp10,
                                     unsigned int width, unsigned int height,
                                     unsigned int src_stride, unsigned int dst_stride) noexcept nogil:
    cdef unsigned int x, y
    cdef unsigned short *b
    cdef unsigned short *g
    cdef unsigned short *r
    cdef unsigned int *dst
    for y in range(height):
        dst = <unsigned int*> (r210 + y*dst_stride)
        g = <unsigned short*> ((<uintptr_t> gbrp10[0]) + y*src_stride)
        b = <unsigned short*> ((<uintptr_t> gbrp10[1]) + y*src_stride)
        r = <unsigned short*> ((<uintptr_t> gbrp10[2]) + y*src_stride)
        for x in range(width):
            dst[x] = (b[x] & 0x3ff) + ((g[x] & 0x3ff)<<10) + ((r[x] & 0x3ff)<<20)


cdef inline void r210_to_YUV444P10_copy(unsigned short *Y, unsigned short *U, unsigned short *V, uintptr_t r210data,
                                        unsigned int width, unsigned int height,
                                        unsigned int Ystride, unsigned int Ustride, unsigned int Vstride,
                                        unsigned int r210_stride) noexcept nogil:
    cdef const unsigned int *r210_row
    cdef unsigned int r210
    cdef unsigned int R, G, B
    cdef unsigned int x, y
    for y in range(height):
        r210_row = <unsigned int*> (r210data + r210_stride*y)
        for x in range(width):
            r210 = r210_row[x]
            R = (r210&0x3ff00000) >> 20
            G = (r210&0x000ffc00) >> 10
            B = (r210&0x000003ff)
            Y[x] = clamp10(YR * R + YG * G + YB * B + YC*4)
            U[x] = clamp10(UR * R + UG * G + UB * B + UC*4)
            V[x] = clamp10(VR * R + VG * G + VB * B + VC*4)
        Y = <unsigned short *> ((<uintptr_t> Y) + Ystride)
        U = <unsigned short *> ((<uintptr_t> U) + Ustride)
        V = <unsigned short *> ((<uintptr_t> V) + Vstride)


cdef inline void YUV444P10_to_r210_copy(uintptr_t r210data, const unsigned short *Ybuf, const unsigned short *Ubuf, const unsigned short *Vbuf,
                                        unsigned int width, unsigned int height,
                                        unsigned int r210_stride,
                                        unsigned int Ystride, unsigned int Ustride, unsigned int Vstride) noexcept nogil:
        cdef unsigned short *Yrow
        cdef unsigned short *Urow
        cdef unsigned short *Vrow
        cdef short Y, U, V
        cdef unsigned int *r210row
        cdef unsigned int x, y
        for y in range(height):
            Yrow = <unsigned short*> ((<uintptr_t> Ybuf) + y*Ystride)
            Urow = <unsigned short*> ((<uintptr_t> Ubuf) + y*Ustride)
            Vrow = <unsigned short*> ((<uintptr_t> Vbuf) + y*Vstride)
            r210row = <unsigned int*> (r210data + y*r210_stride)
            for x in range(width):
                Y = (Yrow[x] & 0x3ff) - Yc*4
                U = (Urow[x] & 0x3ff) - Uc*4
                V = (Vrow[x] & 0x3ff) - Vc*4
                r210row[x] = (
                    (clamp10(RY * Y + RU * U + RV * V)<<20) |
                    (clamp10(GY * Y + GU * U + GV * V)<<10) |
                    (clamp10(BY * Y + BU * U + BV * V))
                    )


cdef inline void YUV444P_to_BGRX_copy(uintptr_t bgrxdata,
                                     const unsigned char *Ybuf, const unsigned char *Ubuf, const unsigned char *Vbuf,
                                     unsigned int width, unsigned int height,
                                     unsigned int rgb_stride,
                                     unsigned int Ystride, unsigned int Ustride, unsigned int Vstride,
                                     unsigned char full_range) noexcept nogil:
        cdef unsigned char *Yrow
        cdef unsigned char *Urow
        cdef unsigned char *Vrow
        cdef short Y, U, V
        cdef unsigned int *bgrxrow
        cdef unsigned int x, y
        if full_range:
            for y in range(height):
                Yrow = <unsigned char*> ((<uintptr_t> Ybuf) + y*Ystride)
                Urow = <unsigned char*> ((<uintptr_t> Ubuf) + y*Ustride)
                Vrow = <unsigned char*> ((<uintptr_t> Vbuf) + y*Vstride)
                bgrxrow = <unsigned int*> (bgrxdata + y*rgb_stride)
                for x in range(width):
                    Y = Yrow[x] - Yc
                    U = Urow[x] - Uc
                    V = Vrow[x] - Vc
                    bgrxrow[x] = (
                        (<unsigned int> (0xff)<<24) |
                        (clamp(RY * Y + RU * U + RV * V)<<16) |
                        (clamp(GY * Y + GU * U + GV * V)<<8) |
                        (clamp(BY * Y + BU * U + BV * V))
                    )
        else:
            for y in range(height):
                Yrow = <unsigned char*> ((<uintptr_t> Ybuf) + y*Ystride)
                Urow = <unsigned char*> ((<uintptr_t> Ubuf) + y*Ustride)
                Vrow = <unsigned char*> ((<uintptr_t> Vbuf) + y*Vstride)
                bgrxrow = <unsigned int*> (bgrxdata + y*rgb_stride)
                for x in range(width):
                    Y = Yrow[x] - SYc
                    U = Urow[x] - SUc
                    V = Vrow[x] - SVc
                    bgrxrow[x] = (
                        (<unsigned int> (0xff)<<24) |
                        (clamp(SRY * Y + SRU * U + SRV * V)<<16) |
                        (clamp(SGY * Y + SGU * U + SGV * V)<<8) |
                        (clamp(SBY * Y + SBU * U + SBV * V))
                    )


cdef inline void YUV444P_to_RGBX_copy(uintptr_t rgbxdata,
                                     const unsigned char *Ybuf, const unsigned char *Ubuf, const unsigned char *Vbuf,
                                     unsigned int width, unsigned int height,
                                     unsigned int rgb_stride,
                                     unsigned int Ystride, unsigned int Ustride, unsigned int Vstride,
                                     unsigned char full_range) noexcept nogil:
        cdef unsigned char *Yrow
        cdef unsigned char *Urow
        cdef unsigned char *Vrow
        cdef short Y, U, V
        cdef unsigned int *bgrxrow
        cdef unsigned int x, y
        if full_range:
            for y in range(height):
                Yrow = <unsigned char*> ((<uintptr_t> Ybuf) + y*Ystride)
                Urow = <unsigned char*> ((<uintptr_t> Ubuf) + y*Ustride)
                Vrow = <unsigned char*> ((<uintptr_t> Vbuf) + y*Vstride)
                rgbxrow = <unsigned int*> (rgbxdata + y*rgb_stride)
                for x in range(width):
                    Y = Yrow[x] - Yc
                    U = Urow[x] - Uc
                    V = Vrow[x] - Vc
                    rgbxrow[x] = (
                        (<unsigned int> 0xff<<24) |
                        (clamp(BY * Y + GU * U + GV * V)<<8) |
                        (clamp(GY * Y + RU * U + RV * V)<<16) |
                        (clamp(RY * Y + BU * U + BV * V))
                    )
        else:
            for y in range(height):
                Yrow = <unsigned char*> ((<uintptr_t> Ybuf) + y*Ystride)
                Urow = <unsigned char*> ((<uintptr_t> Ubuf) + y*Ustride)
                Vrow = <unsigned char*> ((<uintptr_t> Vbuf) + y*Vstride)
                rgbxrow = <unsigned int*> (rgbxdata + y*rgb_stride)
                for x in range(width):
                    Y = Yrow[x] - SYc
                    U = Urow[x] - SUc
                    V = Vrow[x] - SVc
                    rgbxrow[x] = (
                        (<unsigned int> 0xff<<24) |
                        (clamp(SBY * Y + SGU * U + SGV * V)<<8) |
                        (clamp(SGY * Y + SRU * U + SRV * V)<<16) |
                        (clamp(SRY * Y + SBU * U + SBV * V))
                    )


cdef class Converter:
    cdef unsigned int src_width
    cdef unsigned int src_height
    cdef object src_format
    cdef unsigned int dst_width
    cdef unsigned int dst_height
    cdef object dst_format
    cdef unsigned long[3] dst_strides
    cdef unsigned long[3] dst_sizes
    cdef unsigned long[3] offsets
    cdef unsigned char full_range

    cdef convert_image_function

    cdef unsigned long frames
    cdef double time
    cdef unsigned long buffer_size

    cdef object __weakref__

    def init_context(self, int src_width, int src_height, src_format: str,
                           int dst_width, int dst_height, dst_format: str, options: typedict) -> None:
        cdef int i
        assert src_format in COLORSPACES, "invalid input colorspace: %s (must be one of %s)" % (src_format, tuple(COLORSPACES.keys()))
        assert dst_format in COLORSPACES[src_format], "invalid output colorspace: %s (must be one of %s)" % (dst_format, COLORSPACES[src_format])
        log("csc_cython.Converter.init_context%s", (
            src_width, src_height, src_format, dst_width, dst_height, dst_format, options))
        self.src_width = src_width
        self.src_height = src_height
        self.dst_width = dst_width
        self.dst_height = dst_height
        self.src_format = src_format
        self.dst_format = dst_format
        self.full_range = options.boolget("full-range", True)

        self.time = 0
        self.frames = 0

        #explicitly clear all strides / sizes / offsets:
        for i in range(2):
            self.dst_strides[i] = 0
            self.dst_sizes[i]   = 0
            self.offsets[i]     = 0

        def assert_no_scaling() -> None:
            assert src_width==dst_width and src_height==dst_height, "scaling is not supported for %s to %s" % (src_format, dst_format)

        def allocate_yuv(fmt="YUV420P", Bpp=1) -> None:
            divs = get_subsampling_divs(fmt)
            assert divs, "invalid pixel format '%s'" % fmt
            for i, div in enumerate(divs):
                xdiv, ydiv = div
                self.dst_strides[i] = roundup(roundup(self.dst_width*Bpp, xdiv)//xdiv, STRIDE_ROUNDUP)
                self.dst_sizes[i] = self.dst_strides[i] * roundup(self.dst_height, ydiv)//ydiv
            #U channel follows Y with 1 line padding, V follows U with another line of padding:
            self.offsets[0] = 0
            self.offsets[1] = self.offsets[0] + self.dst_sizes[0] + self.dst_strides[0]
            self.offsets[2] = self.offsets[1] + self.dst_sizes[1] + self.dst_strides[1]
            #output buffer ends after V + 1 line of padding:
            self.buffer_size = self.offsets[2] + self.dst_sizes[2] + self.dst_strides[2]
            log("allocate_yuv(%s, %i) buffer_size=%s, sizes=%s, strides=%s",
                fmt, Bpp, self.buffer_size,
                (self.dst_sizes[0], self.dst_sizes[1], self.dst_sizes[2]),
                (self.dst_strides[0], self.dst_strides[1], self.dst_strides[2])
                )

        def allocate_rgb(Bpp=4) -> None:
            self.dst_strides[0] = roundup(self.dst_width*Bpp, STRIDE_ROUNDUP)
            self.dst_sizes[0] = self.dst_strides[0] * self.dst_height
            self.buffer_size = self.dst_sizes[0]+self.dst_strides[0]
            log("allocate_rgb(%i) buffer_size=%s, dst size=%s, stride=%s",
                Bpp, self.buffer_size, self.dst_sizes[0], self.dst_strides[0])

        if src_format in ("BGRX", "RGBX", "RGB", "BGR", "r210") and dst_format in ("YUV420P", "YUV444P"):
            if dst_format == "YUV420P":
                allocate_yuv(dst_format)
                if src_format=="BGRX":
                    self.convert_image_function = self.BGRX_to_YUV420P
                elif src_format=="RGBX":
                    self.convert_image_function = self.RGBX_to_YUV420P
                elif src_format=="BGR":
                    self.convert_image_function = self.BGR_to_YUV420P
                elif src_format=="RGB":
                    self.convert_image_function = self.RGB_to_YUV420P
                else:
                    assert src_format=="r210"
                    self.convert_image_function = self.r210_to_YUV420P
            elif dst_format == "YUV444P":
                allocate_yuv(dst_format)
                if src_format=="BGRX":
                    self.convert_image_function = self.BGRX_to_YUV444P
                elif src_format=="RGBX":
                    self.convert_image_function = self.RGBX_to_YUV444P
                elif src_format=="BGR":
                    self.convert_image_function = self.BGR_to_YUV444P
                elif src_format=="RGB":
                    self.convert_image_function = self.RGB_to_YUV444P
                else:
                    raise ValueError(f"unsupported dst format {dst_format} for {src_format}")
            else:
                raise ValueError(f"unsupported dst format {dst_format}")
        elif src_format=="r210" and dst_format=="BGR48":
            assert_no_scaling()
            allocate_rgb(6)
            self.convert_image_function = self.r210_to_BGR48
        elif src_format=="r210" and dst_format=="YUV444P10":
            assert_no_scaling()
            allocate_yuv(dst_format, 2)
            self.convert_image_function = self.r210_to_YUV444P10
        elif src_format=="YUV444P10" and dst_format=="r210":
            assert_no_scaling()
            allocate_rgb(4)
            self.convert_image_function = self.YUV444P10_to_r210
        elif src_format=="YUV444P" and (dst_format=="BGRX" or dst_format=="RGBX"):
            assert_no_scaling()
            allocate_rgb(3)
            self.convert_image_function = self.YUV444P_to_RGB
        elif src_format=="GBRP10" and dst_format=="r210":
            assert_no_scaling()
            allocate_rgb(4)
            self.convert_image_function = self.GBRP10_to_r210
        elif src_format=="YUV420P" and dst_format in ("RGBX", "BGRX", "RGB", "BGR"):
            #3 or 4 bytes per pixel:
            allocate_rgb(len(dst_format))
            if dst_format=="RGBX":
                self.convert_image_function = self.YUV420P_to_RGBX
            elif dst_format=="BGRX":
                self.convert_image_function = self.YUV420P_to_BGRX
            elif dst_format=="RGB":
                self.convert_image_function = self.YUV420P_to_RGB
            else:
                assert dst_format=="BGR"
                self.convert_image_function = self.YUV420P_to_BGR
        elif src_format=="GBRP" and dst_format in ("RGBX", "BGRX"):
            #4 bytes per pixel:
            allocate_rgb(4)
            if dst_format=="RGBX":
                self.convert_image_function = self.GBRP_to_RGBX
            else:
                assert dst_format=="BGRX"
                self.convert_image_function = self.GBRP_to_BGRX
        else:
            raise ValueError("BUG: src_format=%s, dst_format=%s", src_format, dst_format)

    def clean(self) -> None:
        # overzealous clean is cheap!
        cdef int i                          #
        self.src_width = 0
        self.src_height = 0
        self.dst_width = 0
        self.dst_height = 0
        self.src_format = ""
        self.dst_format = ""
        self.time = 0
        self.frames = 0
        for i in range(3):
            self.dst_strides[i] = 0
            self.dst_sizes[i] = 0
            self.offsets[i] = 0
        self.convert_image_function = None
        self.buffer_size = 0

    def is_closed(self) -> bool:
        return self.convert_image_function is None

    def get_info(self) -> Dict[str,Any]:
        info = {
                "frames"    : self.frames,
                "src_width" : self.src_width,
                "src_height": self.src_height,
                "dst_width" : self.dst_width,
                "dst_height": self.dst_height,
                }
        if self.src_format:
            info["src_format"] = self.src_format
        if self.dst_format:
            info["dst_format"] = self.dst_format
        if self.frames>0 and self.time>0:
            pps = float(self.src_width) * float(self.src_height) * float(self.frames) / self.time
            info["total_time_ms"] = int(self.time * 1000)
            info["pixels_per_second"] = int(pps)
        return info

    def __repr__(self):
        return "csc_cython(%s %sx%s - %s %sx%s)" % (self.src_format, self.src_width, self.src_height,
                                                 self.dst_format, self.dst_width, self.dst_height)

    def __dealloc__(self):
        self.clean()

    def get_src_width(self) -> int:
        return self.src_width

    def get_src_height(self) -> int:
        return self.src_height

    def get_src_format(self) -> str:
        return self.src_format

    def get_dst_width(self) -> int:
        return self.dst_width

    def get_dst_height(self) -> int:
        return self.dst_height

    def get_dst_format(self) -> str:
        return self.dst_format

    def get_type(self) -> str:
        return  "cython"


    def convert_image(self, image: ImageWrapper) -> CythonImageWrapper:
        cdef double start = time.time()
        fn = self.convert_image_function
        r = fn(image)
        cdef double elapsed = time.time()-start
        log("%s took %.1fms", fn, 1000.0*elapsed)
        self.time += elapsed
        self.frames += 1
        return r


    def r210_to_YUV420P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV420P(image, 4, 0, 0, 0)

    def BGR_to_YUV420P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV420P(image, 3, BGR_R, BGR_G, BGR_B)

    def RGB_to_YUV420P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV420P(image, 3, RGB_B, RGB_G, RGB_R)

    def BGRX_to_YUV420P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV420P(image, 4, BGRX_R, BGRX_G, BGRX_B)

    def RGBX_to_YUV420P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV420P(image, 4, RGBX_B, RGBX_G, RGBX_R)

    cdef do_RGB_to_YUV420P(self,
                           image: ImageWrapper,
                           const uint8_t Bpp,
                           const uint8_t Rindex,
                           const uint8_t Gindex,
                           const uint8_t Bindex,
                           ):
        cdef const unsigned int *input_r210
        cdef unsigned int x,y,o
        cdef unsigned int sx, sy, ox, oy
        cdef unsigned unsigned int r210
        cdef unsigned char R, G, B
        cdef unsigned short Rsum, Gsum, Bsum
        cdef unsigned char count, dx, dy

        self.validate_rgb_image(image)
        pixels = image.get_pixels()
        cdef unsigned int input_stride = image.get_rowstride()
        log("do_RGB_to_YUV420P(%s, %i, %i, %i, %i) input=%s, strides=%s", image, Bpp, Rindex, Gindex, Bindex, len(pixels), input_stride)

        #allocate output buffer:
        cdef unsigned char *output_image = <unsigned char*> memalign(self.buffer_size)
        cdef unsigned char *Y = output_image + self.offsets[0]
        cdef unsigned char *U = output_image + self.offsets[1]
        cdef unsigned char *V = output_image + self.offsets[2]

        #copy to local variables (ensures C code will be optimized correctly)
        cdef unsigned int Ystride = self.dst_strides[0]
        cdef unsigned int Ustride = self.dst_strides[1]
        cdef unsigned int Vstride = self.dst_strides[2]
        cdef unsigned int src_width = self.src_width
        cdef unsigned int src_height = self.src_height
        cdef unsigned int dst_width = self.dst_width
        cdef unsigned int dst_height = self.dst_height
        cdef unsigned char full_range = self.full_range

        #we process 4 pixels at a time:
        cdef unsigned int workw = roundup(dst_width, 2)//2
        cdef unsigned int workh = roundup(dst_height, 2)//2

        cdef Py_buffer py_buf
        if PyObject_GetBuffer(pixels, &py_buf, PyBUF_ANY_CONTIGUOUS):
            raise ValueError("failed to read pixel data from %s" % type(pixels))
        cdef const unsigned char *input_image = <const unsigned char *> py_buf.buf

        #from now on, we can release the gil:
        if self.src_format=="r210":
            assert Bpp==4
            input_r210 = <unsigned int*> input_image
            with nogil:
                for y in range(workh):
                    for x in range(workw):
                        Rsum = Gsum = Bsum = 0
                        count = 0
                        for dy in range(2):
                            oy = y*2 + dy
                            if oy>=dst_height:
                                break
                            sy = oy*src_height//dst_height
                            for dx in range(2):
                                ox = x*2 + dx
                                if ox>=dst_width:
                                    break
                                sx = ox*src_width//dst_width
                                o = sy*input_stride + sx*Bpp
                                r210 = input_r210[o//4]
                                B = (r210&0x3ff00000) >> 22
                                G = (r210&0x000ffc00) >> 12
                                R = (r210&0x000003ff) >> 2
                                o = oy*Ystride + ox
                                Y[o] = clamp(YR * R + YG * G + YB * B + YC)
                                count += 1
                                Rsum += R
                                Gsum += G
                                Bsum += B
                        #write 1U and 1V:
                        if count>0:
                            U[y*Ustride + x] = clamp(UR * Rsum//count + UG * Gsum//count + UB * Bsum//count + UC)
                            V[y*Vstride + x] = clamp(VR * Rsum//count + VG * Gsum//count + VB * Bsum//count + VC)
        else:
            with nogil:
                for y in range(workh):
                    for x in range(workw):
                        Rsum = Gsum = Bsum = 0
                        count = 0
                        for dy in range(2):
                            oy = y*2 + dy
                            if oy>=dst_height:
                                break
                            sy = oy*src_height//dst_height
                            for dx in range(2):
                                ox = x*2 + dx
                                if ox>=dst_width:
                                    break
                                sx = ox*src_width//dst_width
                                o = sy*input_stride + sx*Bpp
                                R = input_image[o + Rindex]
                                G = input_image[o + Gindex]
                                B = input_image[o + Bindex]
                                o = oy*Ystride + ox
                                if full_range:
                                    Y[o] = clamp(YR * R + YG * G + YB * B + YC)
                                else:
                                    Y[o] = clamp_studio_Y(YR * R + YG * G + YB * B + YC)
                                count += 1
                                Rsum += R
                                Gsum += G
                                Bsum += B
                        #write 1U and 1V:
                        if count>0:
                            Rsum /= count
                            Gsum /= count
                            Bsum /= count
                            if full_range:
                                U[y*Ustride + x] = clamp(UR * Rsum + UG * Gsum + UB * Bsum + UC)
                                V[y*Vstride + x] = clamp(VR * Rsum + VG * Gsum + VB * Bsum + VC)
                            else:
                                U[y*Ustride + x] = clamp_studio_UV(UR * Rsum + UG * Gsum + UB * Bsum + UC)
                                V[y*Vstride + x] = clamp_studio_UV(VR * Rsum + VG * Gsum + VB * Bsum + VC)
        PyBuffer_Release(&py_buf)
        return self.planar3_image_wrapper(<void *> output_image)

    def BGR_to_YUV444P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV444P(image, 3, BGR_R, BGR_G, BGR_B)

    def RGB_to_YUV444P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV444P(image, 3, RGB_B, RGB_G, RGB_R)

    def BGRX_to_YUV444P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV444P(image, 4, BGRX_R, BGRX_G, BGRX_B)

    def RGBX_to_YUV444P(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGB_to_YUV444P(image, 4, RGBX_B, RGBX_G, RGBX_R)

    cdef do_RGB_to_YUV444P(self,
                           image: ImageWrapper,
                           const uint8_t Bpp,
                           const uint8_t Rindex,
                           const uint8_t Gindex,
                           const uint8_t Bindex,
                           ):
        cdef unsigned int x, y, o
        cdef unsigned int sx, sy
        cdef unsigned char R, G, B

        self.validate_rgb_image(image)
        pixels = image.get_pixels()
        cdef unsigned int input_stride = image.get_rowstride()
        log("do_RGB_to_YUV444P(%s, %i, %i, %i, %i) input=%s, strides=%s", image, Bpp, Rindex, Gindex, Bindex, len(pixels), input_stride)

        #allocate output buffer:
        cdef unsigned char *output_image = <unsigned char*> memalign(self.buffer_size)
        cdef unsigned char *Y = output_image + self.offsets[0]
        cdef unsigned char *U = output_image + self.offsets[1]
        cdef unsigned char *V = output_image + self.offsets[2]

        #copy to local variables (ensures C code will be optimized correctly)
        cdef unsigned int Ystride = self.dst_strides[0]
        cdef unsigned int Ustride = self.dst_strides[1]
        cdef unsigned int Vstride = self.dst_strides[2]
        cdef unsigned int src_width = self.src_width
        cdef unsigned int src_height = self.src_height
        cdef unsigned int dst_width = self.dst_width
        cdef unsigned int dst_height = self.dst_height
        cdef unsigned char full_range = self.full_range

        cdef const unsigned char *input_image
        cdef uintptr_t buf
        with buffer_context(pixels) as bc:
            input_image = <const unsigned char *> (<uintptr_t> int(bc))
            # buf_size = len(bc)
            with nogil:
                for y in range(dst_height):
                    for x in range(dst_width):
                        sx = x * src_width // dst_width
                        sy = y * src_height // dst_height
                        o = sy * input_stride + sx * Bpp
                        # assert o + Bpp <= buf_size
                        R = input_image[o + Rindex]
                        G = input_image[o + Gindex]
                        B = input_image[o + Bindex]
                        if full_range:
                            Y[y * Ystride + x] = clamp(YR * R + YG * G + YB * B + YC)
                            U[y * Ustride + x] = clamp(UR * R + UG * G + UB * B + UC)
                            V[y * Vstride + x] = clamp(VR * R + VG * G + VB * B + VC)
                        else:
                            Y[y * Ystride + x] = clamp_studio_Y(YR * R + YG * G + YB * B + YC)
                            U[y * Ustride + x] = clamp_studio_UV(UR * R + UG * G + UB * B + UC)
                            V[y * Vstride + x] = clamp_studio_UV(VR * R + VG * G + VB * B + VC)
        return self.planar3_image_wrapper(<void *> output_image)

    cdef planar3_image_wrapper(self, void *buf, unsigned char bpp=24):
        #create python buffer from each plane:
        planes = []
        strides = []
        cdef unsigned char i
        for i in range(3):
            strides.append(self.dst_strides[i])
            planes.append(PyMemoryView_FromMemory(<char *> ((<uintptr_t> buf) + self.offsets[i]), self.dst_sizes[i], PyBUF_WRITE))
        out_image = CythonImageWrapper(0, 0, self.dst_width, self.dst_height,
                                       planes, self.dst_format, bpp, strides,
                                       planes=ImageWrapper.PLANAR_3)
        out_image.set_full_range(self.full_range)
        out_image.cython_buffer = <uintptr_t> buf
        return out_image

    def validate_rgb_image(self, image: ImageWrapper) -> None:
        assert image.get_planes()==ImageWrapper.PACKED, "invalid input format: %s planes" % image.get_planes()
        assert image.get_width()>=self.src_width, "invalid image width: %s (minimum is %s)" % (image.get_width(), self.src_width)
        assert image.get_height()>=self.src_height, "invalid image height: %s (minimum is %s)" % (image.get_height(), self.src_height)
        assert image.get_pixels(), "failed to get pixels from %s" % image

    def r210_to_YUV444P10(self, image: ImageWrapper) -> CythonImageWrapper:
        self.validate_rgb_image(image)
        pixels = image.get_pixels()
        cdef unsigned int input_stride = image.get_rowstride()
        log("r210_to_YUV444P10(%s) input=%s, strides=%s", image, len(pixels), input_stride)

        #allocate output buffer:
        cdef void *output_image = memalign(self.buffer_size)
        cdef unsigned short *Y = <unsigned short *> ((<uintptr_t> output_image) + self.offsets[0])
        cdef unsigned short *U = <unsigned short *> ((<uintptr_t> output_image) + self.offsets[1])
        cdef unsigned short *V = <unsigned short *> ((<uintptr_t> output_image) + self.offsets[2])
        #copy to local variables (ensures C code will be optimized correctly)
        cdef unsigned int Ystride = self.dst_strides[0]
        cdef unsigned int Ustride = self.dst_strides[1]
        cdef unsigned int Vstride = self.dst_strides[2]

        cdef uintptr_t buf
        with buffer_context(pixels) as bc:
            buf = <uintptr_t> int(bc)
            if image.is_thread_safe():
                with nogil:
                    r210_to_YUV444P10_copy(Y, U, V, buf,
                                           self.dst_width, self.dst_height,
                                           Ystride, Ustride, Vstride,
                                           input_stride)
            else:
                r210_to_YUV444P10_copy(Y, U, V, buf,
                                       self.dst_width, self.dst_height,
                                       Ystride, Ustride, Vstride,
                                       input_stride)
        return self.planar3_image_wrapper(output_image)

    def YUV444P10_to_r210(self, image: ImageWrapper) -> CythonImageWrapper:
        self.validate_planar3_image(image)
        planes = image.get_pixels()
        input_strides = image.get_rowstride()
        log("YUV444P10_to_r210(%s) strides=%s", image, input_strides)

        #copy to local variables:
        cdef unsigned int width = self.dst_width
        cdef unsigned int height = self.dst_height

        #allocate output buffer:
        cdef char *output_image = <char *> memalign(self.buffer_size)
        cdef unsigned int stride = self.dst_strides[0]

        cdef unsigned int YUVstrides[3]
        cdef Py_buffer py_buf[3]
        cdef const unsigned short * YUV[3]
        cdef int i
        for i in range(3):
            YUVstrides[i] = input_strides[i]
            if PyObject_GetBuffer(planes[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                raise ValueError("failed to read pixel data from %s" % type(planes[i]))
            YUV[i] = <const unsigned short *> py_buf[i].buf
            min_len = YUVstrides[i]*image.get_height()
            assert py_buf.len>=min_len, "buffer for Y plane is too small: %s bytes, expected at least %s" % (py_buf.len, min_len)

        if image.is_thread_safe():
            with nogil:
                YUV444P10_to_r210_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                       width, height,
                                       stride,
                                       YUVstrides[0], YUVstrides[1], YUVstrides[2])
        else:
            YUV444P10_to_r210_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                   width, height,
                                   stride,
                                   YUVstrides[0], YUVstrides[1], YUVstrides[2])
        for i in range(3):
            PyBuffer_Release(&py_buf[i])
        return self.packed_image_wrapper(output_image, 30)

    def YUV444P_to_RGB(self, image: ImageWrapper) -> CythonImageWrapper:
        self.validate_planar3_image(image)
        planes = image.get_pixels()
        input_strides = image.get_rowstride()
        log("YUV444P_to_RGB(%s) strides=%s", image, input_strides)

        #copy to local variables:
        cdef unsigned int width = self.dst_width
        cdef unsigned int height = self.dst_height

        #allocate output buffer:
        cdef char *output_image = <char *> memalign(self.buffer_size)
        cdef unsigned int stride = self.dst_strides[0]
        cdef unsigned char full_range = self.full_range

        cdef unsigned int YUVstrides[3]
        cdef Py_buffer py_buf[3]
        cdef const unsigned char * YUV[3]
        cdef int i
        for i in range(3):
            YUVstrides[i] = input_strides[i]
            if PyObject_GetBuffer(planes[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                raise ValueError("failed to read pixel data from %s" % type(planes[i]))
            YUV[i] = <const unsigned char *> py_buf[i].buf
            min_len = YUVstrides[i]*image.get_height()
            assert py_buf.len>=min_len, "buffer for Y plane is too small: %s bytes, expected at least %s" % (py_buf.len, min_len)

        assert self.dst_format=="BGRX" or self.dst_format=="RGBX"
        if image.is_thread_safe():
            if self.dst_format=="BGRX":
                with nogil:
                    YUV444P_to_BGRX_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                         width, height,
                                         stride,
                                         YUVstrides[0], YUVstrides[1], YUVstrides[2],
                                         full_range)
            else:
                with nogil:
                    YUV444P_to_RGBX_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                         width, height,
                                         stride,
                                         YUVstrides[0], YUVstrides[1], YUVstrides[2],
                                         full_range)
        else:
            if self.dst_format=="BGRX":
                YUV444P_to_BGRX_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                     width, height,
                                     stride,
                                     YUVstrides[0], YUVstrides[1], YUVstrides[2],
                                     full_range)
            else:
                YUV444P_to_RGBX_copy(<uintptr_t> output_image, YUV[0], YUV[1], YUV[2],
                                     width, height,
                                     stride,
                                     YUVstrides[0], YUVstrides[1], YUVstrides[2],
                                     full_range)
        for i in range(3):
            PyBuffer_Release(&py_buf[i])
        return self.packed_image_wrapper(output_image, 30)

    def r210_to_BGR48(self, image: ImageWrapper) -> CythonImageWrapper:
        self.validate_rgb_image(image)
        pixels = image.get_pixels()
        input_stride = image.get_rowstride()
        log("r210_to_BGR48(%s) input=%s, strides=%s", image, len(pixels), input_stride)

        #allocate output buffer:
        cdef unsigned short *bgr48 = <unsigned short*> memalign(self.dst_sizes[0])

        cdef unsigned int w = self.src_width
        cdef unsigned int h = self.src_height
        cdef unsigned int src_stride = image.get_rowstride()
        cdef unsigned int dst_stride = self.dst_strides[0]

        assert (dst_stride%2)==0

        cdef const unsigned int *r210
        with buffer_context(pixels) as bc:
            r210 = <const unsigned int*> (<uintptr_t> int(bc))
            if image.is_thread_safe():
                with nogil:
                    r210_to_BGR48_copy(bgr48, r210, w, h, src_stride, dst_stride)
            else:
                r210_to_BGR48_copy(bgr48, r210, w, h, src_stride, dst_stride)
        return self.packed_image_wrapper(<char *> bgr48, 48)

    cdef packed_image_wrapper(self, char *buf, unsigned char bpp=24):
        pybuf = PyMemoryView_FromMemory(buf, self.dst_sizes[0], PyBUF_WRITE)
        out_image = CythonImageWrapper(0, 0, self.dst_width, self.dst_height, pybuf, self.dst_format, bpp, self.dst_strides[0], planes=ImageWrapper.PACKED)
        out_image.cython_buffer = <uintptr_t> buf
        return out_image

    def validate_planar3_image(self, image: ImageWrapper) -> None:
        assert image.get_planes()==ImageWrapper.PLANAR_3, "invalid input format: %s planes" % image.get_planes()
        assert image.get_width()>=self.src_width, "invalid image width: %s (minimum is %s)" % (image.get_width(), self.src_width)
        assert image.get_height()>=self.src_height, "invalid image height: %s (minimum is %s)" % (image.get_height(), self.src_height)
        assert image.get_pixels(), "failed to get pixels from %s" % image

    def GBRP10_to_r210(self, image: ImageWrapper) -> CythonImageWrapper:
        self.validate_planar3_image(image)
        pixels = image.get_pixels()
        input_strides = image.get_rowstride()
        cdef unsigned int w = self.src_width
        cdef unsigned int h = self.src_height
        cdef unsigned int src_stride = input_strides[0]
        cdef unsigned int dst_stride = self.dst_strides[0]
        assert input_strides[1]==src_stride and input_strides[2]==src_stride, "mismatch in rowstrides: %s" % (input_strides,)
        assert src_stride>=w*2
        assert dst_stride>=w*4
        assert self.dst_sizes[0]>=dst_stride*h

        #allocate output buffer:
        cdef unsigned int *r210 = <unsigned int*> memalign(self.dst_sizes[0])

        cdef uintptr_t gbrp10[3]
        cdef unsigned int i
        cdef Py_buffer py_buf[3]
        for i in range(3):
            if PyObject_GetBuffer(pixels[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                raise ValueError("failed to read pixel data from %s" % type(pixels[i]))
            gbrp10[i] = <uintptr_t> py_buf[i].buf
            assert (<unsigned long> py_buf[i].len)>=src_stride*h, "input plane '%s' is too small: %i bytes" % ("GBR"[i], py_buf[i].len)

        if image.is_thread_safe():
            with nogil:
                gbrp10_to_r210_copy(<uintptr_t> r210, gbrp10,
                                    w, h,
                                    src_stride, dst_stride)
        else:
            gbrp10_to_r210_copy(<uintptr_t> r210, gbrp10,
                                w, h,
                                src_stride, dst_stride)
        for i in range(3):
            PyBuffer_Release(&py_buf[i])
        return self.packed_image_wrapper(<char *> r210, 30)

    def YUV420P_to_RGBX(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_YUV420P_to_RGB(image, 4, RGBX_B, RGBX_G, RGBX_R, RGBX_X)

    def YUV420P_to_RGB(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_YUV420P_to_RGB(image, 3, RGB_B, RGB_G, RGB_R, 0)

    def YUV420P_to_BGRX(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_YUV420P_to_RGB(image, 4, BGRX_R, BGRX_G, BGRX_B, BGRX_X)

    def YUV420P_to_BGR(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_YUV420P_to_RGB(image, 3, BGR_R, BGR_G, BGR_B, 0)

    cdef do_YUV420P_to_RGB(self, image: ImageWrapper,
                           const uint8_t Bpp,
                           const uint8_t Rindex,
                           const uint8_t Gindex,
                           const uint8_t Bindex,
                           const uint8_t Xindex,
                           ):
        cdef unsigned int x,y,o
        cdef unsigned int sx, sy, ox, oy
        cdef unsigned char dx, dy
        cdef short Y, U, V

        self.validate_planar3_image(image)
        planes = image.get_pixels()
        input_strides = image.get_rowstride()
        log("do_YUV420P_to_RGB(%s) strides=%s", (image, Bpp, Rindex, Gindex, Bindex, Xindex), input_strides)

        #allocate output buffer:
        cdef unsigned char *output_image = <unsigned char*> memalign(self.buffer_size)

        #copy to local variables:
        cdef unsigned int stride = self.dst_strides[0]
        cdef unsigned int src_width = self.src_width
        cdef unsigned int src_height = self.src_height
        cdef unsigned int dst_width = self.dst_width
        cdef unsigned int dst_height = self.dst_height
        cdef unsigned int Ystride = input_strides[0]
        cdef unsigned int Ustride = input_strides[1]
        cdef unsigned int Vstride = input_strides[2]
        cdef Py_buffer py_buf[3]
        cdef int i
        for i in range(3):
            if PyObject_GetBuffer(planes[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                raise ValueError("failed to read pixel data from %s" % type(planes[i]))
            min_len = input_strides[i]*image.get_height()
            assert py_buf.len>=min_len, "buffer for Y plane is too small: %s bytes, expected at least %s" % (py_buf.len, min_len)
        cdef unsigned char *Ybuf = <unsigned char *> py_buf[0].buf
        cdef unsigned char *Ubuf = <unsigned char *> py_buf[1].buf
        cdef unsigned char *Vbuf = <unsigned char *> py_buf[2].buf
        cdef unsigned char full_range = self.full_range

        cdef unsigned char yc = Yc
        cdef unsigned char uc = Uc
        cdef unsigned char vc = Vc
        cdef int ry = RY, ru = RU, rv = RV
        cdef int gy = GY, gu = GU, gv = GV
        cdef int by = BY, bu = BU, bv = BV
        if not full_range:
            yc = SYc
            uc = SUc
            vc = SVc
            ry = SRY
            ru = SRU
            rv = SRV
            gy = SGY
            gu = SGU
            gv = SGV
            by = SBY
            bu = SBU
            bv = SBV

        #we process 4 pixels at a time:
        cdef unsigned int workw = roundup(dst_width//2, 2)
        cdef unsigned int workh = roundup(dst_height//2, 2)
        #from now on, we can release the gil:
        with nogil:
            for y in range(workh):
                for x in range(workw):
                    #assert x*2<=src_width and y*2<=src_height
                    #read U and V for the next 4 pixels:
                    sx = x*src_width//dst_width
                    sy = y*src_height//dst_height
                    U = Ubuf[sy*Ustride + sx] - uc
                    V = Vbuf[sy*Vstride + sx] - vc
                    #now read up to 4 Y values and write an RGBX pixel for each:
                    for dy in range(2):
                        oy = y*2 + dy
                        if oy>=dst_height:
                            break
                        sy = oy*src_height//dst_height
                        for dx in range(2):
                            ox = x*2 + dx
                            if ox>=dst_width:
                                break
                            sx = ox*src_width//dst_width
                            Y = Ybuf[sy*Ystride + sx] - yc
                            o = oy*stride + ox * Bpp
                            output_image[o + Rindex] = clamp(ry * Y + ru * U + rv * V)
                            output_image[o + Gindex] = clamp(gy * Y + gu * U + gv * V)
                            output_image[o + Bindex] = clamp(by * Y + bu * U + bv * V)
                            if Bpp==4:
                                output_image[o + Xindex] = 255
        for i in range(3):
            PyBuffer_Release(&py_buf[i])
        return self.packed_image_wrapper(<char *> output_image, 24)

    def GBRP_to_RGBX(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGBP_to_RGB(image, 2, 0, 1, RGBX_R, RGBX_G, RGBX_B, RGBX_X)

    def GBRP_to_BGRX(self, image: ImageWrapper) -> CythonImageWrapper:
        return self.do_RGBP_to_RGB(image, 2, 0, 1, RGBX_B, RGBX_G, RGBX_R, RGBX_X)

    cdef do_RGBP_to_RGB(self, image: ImageWrapper,
                        const uint8_t Rsrc,
                        const uint8_t Gsrc,
                        const uint8_t Bsrc,
                        const uint8_t Rdst,
                        const uint8_t Gdst,
                        const uint8_t Bdst,
                        const uint8_t Xdst,
                        ):
        cdef unsigned int x,y,o
        cdef unsigned int sx, sy
        cdef unsigned char *Gptr
        cdef unsigned char *Bptr
        cdef unsigned char *Rptr

        self.validate_planar3_image(image)
        planes = image.get_pixels()
        input_strides = image.get_rowstride()
        log("do_RGBP_to_RGB(%s) strides=%s", (image, Rsrc, Gsrc, Bsrc, Rdst, Gdst, Bdst, Xdst), input_strides)

        #allocate output buffer:
        cdef unsigned char *output_image = <unsigned char*> memalign(self.buffer_size)

        #copy to local variables:
        cdef unsigned int Rstride = input_strides[Rsrc]
        cdef unsigned int Gstride = input_strides[Gsrc]
        cdef unsigned int Bstride = input_strides[Bsrc]
        cdef unsigned int stride = self.dst_strides[0]
        cdef unsigned int src_width = self.src_width
        cdef unsigned int src_height = self.src_height
        cdef unsigned int dst_width = self.dst_width
        cdef unsigned int dst_height = self.dst_height
        cdef Py_buffer py_buf[3]
        cdef int i
        for i in range(3):
            if PyObject_GetBuffer(planes[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                raise ValueError("failed to read pixel data from %s" % type(planes[i]))
            min_len = input_strides[i]*image.get_height()
            assert py_buf.len>=min_len, "buffer for G plane is too small: %s bytes, expected at least %s" % (py_buf.len, min_len)
        cdef unsigned char *Gbuf = <unsigned char*> py_buf[Gsrc].buf
        cdef unsigned char *Bbuf = <unsigned char*> py_buf[Bsrc].buf
        cdef unsigned char *Rbuf = <unsigned char*> py_buf[Rsrc].buf

        #from now on, we can release the gil:
        with nogil:
            for y in range(dst_height):
                o = stride*y
                sy = y*src_height/dst_height
                Rptr  = Rbuf + (sy * Rstride)
                Gptr  = Gbuf + (sy * Gstride)
                Bptr  = Bbuf + (sy * Bstride)
                for x in range(dst_width):
                    sx = x*src_width/dst_width
                    output_image[o+Rdst] = Rptr[sx]
                    output_image[o+Gdst] = Gptr[sx]
                    output_image[o+Bdst] = Bptr[sx]
                    output_image[o+Xdst] = 255
                    o += 4
        for i in range(3):
            PyBuffer_Release(&py_buf[i])
        return self.packed_image_wrapper(<char *> output_image, 24)


def selftest(full=False):
    from xpra.codecs.checks import testcsc
    from xpra.codecs.csc_cython import converter
    testcsc(converter, full)



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2022 Nathalie Casati <nat@yuka.ch>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import json
from collections.abc import Sequence

from xpra.auth.sys_auth_base import SysAuthenticator, log

KEYCLOAK_SERVER_URL = os.environ.get("XPRA_KEYCLOAK_SERVER_URL", "https://localhost:8080/auth/")
KEYCLOAK_REALM_NAME = os.environ.get("XPRA_KEYCLOAK_REALM_NAME", "example_realm")
KEYCLOAK_CLIENT_ID = os.environ.get("XPRA_KEYCLOAK_CLIENT_ID", "example_client")
KEYCLOAK_CLIENT_SECRET_KEY = os.environ.get("XPRA_KEYCLOAK_CLIENT_SECRET_KEY", "secret")
KEYCLOAK_REDIRECT_URI = os.environ.get("XPRA_KEYCLOAK_REDIRECT_URI", "http://localhost/login/")
KEYCLOAK_CLAIM_FIELD = os.environ.get("XPRA_KEYCLOAK_CLAIM_FIELD", "")  # field containing groups claim
KEYCLOAK_AUTH_GROUPS = os.environ.get("XPRA_KEYCLOAK_AUTH_GROUPS", "")  # authorized groups
KEYCLOAK_AUTH_CONDITION = os.environ.get("XPRA_KEYCLOAK_AUTH_CONDITION", "and")  # authentication condition
KEYCLOAK_SCOPE = os.environ.get("XPRA_KEYCLOAK_SCOPE", "openid")
KEYCLOAK_GRANT_TYPE = os.environ.get("XPRA_KEYCLOAK_GRANT_TYPE", "authorization_code")


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        self.server_url = kwargs.pop("server_url", KEYCLOAK_SERVER_URL)
        self.realm_name = kwargs.pop("realm_name", KEYCLOAK_REALM_NAME)
        self.client_id = kwargs.pop("client_id", KEYCLOAK_CLIENT_ID)
        self.client_secret_key = kwargs.pop("client_secret_key", KEYCLOAK_CLIENT_SECRET_KEY)
        self.redirect_uri = kwargs.pop("redirect_uri", KEYCLOAK_REDIRECT_URI)
        self.claim_field = kwargs.pop("claim_field", KEYCLOAK_CLAIM_FIELD)
        self.auth_groups = set(kwargs.pop("auth_groups", KEYCLOAK_AUTH_GROUPS).split())
        self.auth_condition = kwargs.pop("auth_condition", KEYCLOAK_AUTH_CONDITION)
        self.scope = kwargs.pop("scope", KEYCLOAK_SCOPE)
        self.grant_type = kwargs.pop("grant_type", KEYCLOAK_GRANT_TYPE)

        # use keycloak as default prompt
        kwargs["prompt"] = kwargs.pop("prompt", "keycloak")

        if self.grant_type != "authorization_code":
            raise ValueError(f"grant type {self.grant_type!r} is not supported, only \"authorization_code\"")

        super().__init__(**kwargs)
        log("keycloak auth: server_url=%s, client_id=%s, realm_name=%s, redirect_uri=%s, scope=%s, grant_type=%s",
            self.server_url, self.client_id, self.realm_name, self.redirect_uri, self.scope, self.grant_type)

        try:
            # noinspection PyPackageRequirements
            from oauthlib.oauth2 import WebApplicationClient
            # Get authorization code
            client = WebApplicationClient(self.client_id)
            authorization_url = self.server_url + 'realms/' + self.realm_name + '/protocol/openid-connect/auth'
            client_uri = client.prepare_request_uri(authorization_url,
                                                    redirect_uri=self.redirect_uri,
                                                    scope=[self.scope],
                                                    )
            self.salt = client_uri.encode("utf8")
        except ImportError as e:  # pragma: no cover
            log("check(..)", exc_info=True)
            log.warn("Warning: cannot use keycloak authentication:")
            log.warn(" %s", e)
            # unsure how to fail the auth at this point so we raise the exception
            raise

    def __repr__(self):
        return "keycloak"

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        assert not self.challenge_sent
        if "keycloak" not in digests:
            log.error("Error: client does not support keycloak authentication")
            return b"", ""
        self.challenge_sent = True
        return self.salt, "keycloak"

    def check(self, response_json: bytes) -> bool:
        log(f"check({response_json!r})")
        assert self.challenge_sent
        if not response_json:
            log.error("Error: keycloak authentication failed")
            log.error(" invalid response received from authorization endpoint")
            return False

        try:
            response = json.loads(response_json)
        except (json.JSONDecodeError, TypeError):
            log.error("Error: keycloak authentication failed")
            log.error(" invalid response received from authorization endpoint")
            log("failed to parse json: %r", response_json, exc_info=True)
            return False
        log(f"json.loads({response_json!r})={response!r}")

        if not isinstance(response, dict):
            log.error("Error: keycloak authentication failed")
            log.error(" invalid response received from authorization endpoint")
            log("response is of type %r but dict type is required", type(response))
            log("failed to load response %r", response)
            return False

        log("check(%r)", response)
        auth_code = response.get("code")
        error = response.get("error")

        if error:
            log.error("Error: keycloak authentication failed")
            log.error("%s: %s", error, response.get("error_description"))
            return False

        if not auth_code:
            log.error("Error: keycloak authentication failed")
            log.error(" invalid response received from authorization endpoint")
            return False

        try:
            # pylint: disable=import-outside-toplevel
            from keycloak import KeycloakOpenID
            from keycloak.exceptions import KeycloakError
        except ImportError as e:
            log("check(..)", exc_info=True)
            log.warn("Warning: cannot use keycloak authentication:")
            log.warn(" %s", e)
            return False

        try:
            # Configure client
            keycloak_openid = KeycloakOpenID(server_url=self.server_url,
                                             client_id=self.client_id,
                                             realm_name=self.realm_name,
                                             client_secret_key=self.client_secret_key)

            # Get well_known
            if hasattr(keycloak_openid, "well_known"):
                # version 1.8 and later:
                # https://github.com/marcospereirampj/python-keycloak/commit/6bfbd0d15fa5981f35e5a6866b3efd62ef0dc968
                config_well_known = keycloak_openid.well_known()
            else:
                config_well_known = keycloak_openid.well_know()
            log("well_known: %r", config_well_known)

            # Get token
            token = keycloak_openid.token(code=auth_code,
                                          grant_type=[self.grant_type],
                                          redirect_uri=self.redirect_uri)

            # Verify token
            access_token = token.get("access_token")
            if not access_token:
                log.error("Error: keycloak authentication failed as access token is missing")
                return False

            token_info = keycloak_openid.introspect(access_token)
            log("token_info: %r", token_info)

            token_state = token_info.get("active")
            if token_state is None:
                log.error("Error: keycloak authentication failed as token state is missing")
                return False

            if token_state is False:
                log.error("Error: keycloak authentication failed as token state not active")
                return False

            if token_state is not True:
                log.error("Error: keycloak authentication failed as token state is invalid")
                return False

            log("keycloak: token is active")

            user_info = keycloak_openid.userinfo(access_token)
            log("user_info: %r", user_info)

            log("claim_field: %r", self.claim_field)
            log("auth_groups: %r", self.auth_groups)

            # if claim_field is defined, check that the auth_groups match the groups_claim condition
            if self.claim_field:
                if not self.auth_groups:
                    log.error("Error: keycloak authentication failed as auth_groups is invalid")
                    return False

                def extract_groups_claim(data_dict, keys):

                    """
                    Extracts a nested value from a dictionary using a list of keys.

                    This function is specifically designed to handle nested dictionaries,
                    such as the ones provided by Keycloak userinfo.

                    For example, userinfo could contain the following structure:

                    "resource_access": {
                        "roles": {
                            "group": [
                                "client-group-1",
                                "client-group-2"
                            ]
                        }
                    }

                    In this scenario, the function traverses the dictionary using the keys
                    ['resource_access', 'roles', 'group'] to extract the list of groups
                    ['client-group-1', 'client-group-2'].

                    Args:
                        data_dict (dict): The dictionary from which to extract the nested value.
                        keys (list): A list of keys that represents the path to the desired value
                                     in the nested dictionary.

                    Returns:
                        The nested value extracted from the dictionary. If the sequence of keys is
                        exhausted, it returns the portion of the dictionary that it has navigated to.
                    """
                    return extract_groups_claim(data_dict[keys.pop(0)], keys) if len(keys) else data_dict

                groups_claim = set(extract_groups_claim(user_info, self.claim_field.split('.')))
                log("groups_claim: %r", groups_claim)

                if self.auth_condition == "or":
                    if not self.auth_groups.intersection(groups_claim):
                        log.error("Error: keycloak authentication failed as groups claim is not satisfied")
                        return False
                elif self.auth_condition == "and":
                    if not self.auth_groups.issubset(groups_claim):
                        log.error("Error: keycloak authentication failed as groups claim is not satisfied")
                        return False
                else:
                    log.error("Error: keycloak authentication failed as auth_condition is invalid")
                    return False

            log("keycloak authentication succeeded")
            return True
        except KeycloakError as e:
            log.error("Error: keycloak authentication failed")
            log.error(" error code %s: %s", e.response_code, e.error_message)
            return False


def main(args) -> int:  # pragma: no cover
    if len(args) != 2:
        print("invalid number of arguments")
        print("usage:")
        print(f"{args[0]} response_json")
        return 1
    response_json = args[1]

    a = Authenticator()
    a.get_challenge(("keycloak", ))

    if not a.check(response_json):
        print("failed")
        return -1
    print("success")
    return 0


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from abc import ABCMeta, abstractmethod


class AuthenticationHandler(metaclass=ABCMeta):

    @abstractmethod
    def get_digest(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def handle(self, challenge: bytes, digest: str, prompt: str):
        raise NotImplementedError()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from xpra.os_util import gi_import
from xpra.util.str_fn import repr_ellipsized, bytestostr
from xpra.util.system import is_X11
from xpra.net.bytestreams import set_socket_timeout
from xpra.net.rfb.const import RFB_KEYNAMES
from xpra.server.rfb.protocol import RFBServerProtocol
from xpra.server.rfb.source import RFBSource
from xpra.server import features
from xpra.scripts.config import str_to_bool, parse_number
from xpra.log import Logger


GLib = gi_import("GLib")

log = Logger("rfb")
mouselog = Logger("rfb", "mouse")
keylog = Logger("rfb", "keyboard")


class RFBServer:
    """
        Adds RFB packet handler to a server.
    """

    def __init__(self):
        self._window_to_id = {}
        self._rfb_upgrade = 0
        self.readonly = False
        self.rfb_buttons = 0
        self.x11_keycodes_for_keysym = {}
        self.X11Keyboard = None
        if is_X11():
            try:
                from xpra.x11.bindings.keyboard import X11KeyboardBindings
                self.X11Keyboard = X11KeyboardBindings()
            except ImportError:
                log("RFBServer", exc_info=True)
                log.warn("Warning: no x11 bindings")
                log.warn(" some RFB keyboard events may be missing")

    def init(self, opts):
        if not str_to_bool(opts.rfb_upgrade):
            self._rfb_upgrade = 0
        else:
            self._rfb_upgrade = parse_number(int, "rfb-upgrade", opts.rfb_upgrade, 0)
        log("init(..) rfb-upgrade=%i", self._rfb_upgrade)

    def _get_rfb_desktop_model(self):
        models = tuple(self._window_to_id.keys())
        if not models:
            log.error("RFB: no window models to export, dropping connection")
            return None
        if len(models) != 1:
            log.error("RFB can only handle a single desktop window, found %i", len(self._window_to_id))
            return None
        return models[0]

    def _get_rfb_desktop_wid(self):
        ids = tuple(self._window_to_id.values())
        if len(ids) != 1:
            log.error("RFB can only handle a single desktop window, found %i", len(self._window_to_id))
            return None
        return ids[0]

    def handle_rfb_connection(self, conn, data=b""):
        model = self._get_rfb_desktop_model()
        log("handle_rfb_connection(%s) model=%s", conn, model)
        if not model:
            log("no desktop model, closing RFB connection")
            conn.close()
            return

        def rfb_protocol_class(conn):
            auths = self.make_authenticators("rfb", {}, conn)
            assert len(auths) <= 1, "rfb does not support multiple authentication modules"
            auth = None
            if len(auths) == 1:
                auth = auths[0]
            log("creating RFB protocol with authentication=%s", auth)
            return RFBServerProtocol(conn, auth,
                                     self.process_rfb_packet, self.get_rfb_pixelformat,
                                     self.session_name or "Xpra Server",
                                     data)

        p = self.do_make_protocol("rfb", conn, {}, rfb_protocol_class)
        log("handle_rfb_connection(%s) protocol=%s", conn, p)
        p.send_protocol_handshake()

    def process_rfb_packet(self, proto, packet):
        # log("RFB packet: '%s'", packet)
        fn_name = "_process_rfb_%s" % bytestostr(packet[0]).replace("-", "_")
        fn = getattr(self, fn_name, None)
        if not fn:
            log.warn("Warning: no RFB handler for %s", fn_name)
            return
        fn(proto, packet)

    def get_rfb_pixelformat(self):
        model = self._get_rfb_desktop_model()
        w, h = model.get_dimensions()
        # w, h, bpp, depth, bigendian, truecolor, rmax, gmax, bmax, rshift, bshift, gshift
        return w, h, 32, 24, False, True, 255, 255, 255, 16, 8, 0

    def _process_rfb_invalid(self, proto, packet):
        self.disconnect_protocol(proto, "invalid packet: %s" % repr_ellipsized(packet[1:]))

    def _process_rfb_connection_lost(self, proto, packet):
        self._process_connection_lost(proto, packet)

    def _process_rfb_authenticated(self, proto, _packet):
        model = self._get_rfb_desktop_model()
        if not model:
            proto.close()
            return
        self.accept_protocol(proto)
        # use blocking sockets from now on:
        set_socket_timeout(proto._conn, None)
        accepted, share_count, disconnected = self.handle_sharing(proto, share=proto.share)
        log("RFB handle sharing: accepted=%s, share count=%s, disconnected=%s", accepted, share_count, disconnected)
        if not accepted:
            return
        source = RFBSource(proto, proto.share)
        self._server_sources[proto] = source
        # continue in the UI thread:
        GLib.idle_add(self._accept_rfb_source, source)

    def _accept_rfb_source(self, source):
        if features.input_devices:
            source.keyboard_config = self.get_keyboard_config()
            self.set_keymap(source)
        model = self._get_rfb_desktop_model()
        w, h = model.get_dimensions()
        source.damage(self._window_to_id[model], model, 0, 0, w, h)
        # ugly weak dependency,
        # shadow servers need to be told to start the refresh timer:
        start_refresh = getattr(self, "start_refresh", None)
        if start_refresh:
            for wid in tuple(self._window_to_id.values()):
                start_refresh(wid)  # pylint: disable=not-callable

    def _process_rfb_PointerEvent(self, _proto, packet):
        if not features.input_devices or self.readonly:
            return
        buttons, x, y = packet[1:4]
        wid = self._get_rfb_desktop_wid()

        def process_pointer_event():
            mouselog("RFB PointerEvent(%#x, %s, %s) desktop wid=%s", buttons, x, y, wid)
            device_id = -1
            self._move_pointer(device_id, wid, (x, y))
            if buttons != self.rfb_buttons:
                # figure out which buttons have changed:
                for button in range(8):
                    mask = 2 ** button
                    if buttons & mask != self.rfb_buttons & mask:
                        pressed = bool(buttons & mask)
                        mouselog(" %spressing button %i", ["un", ""][pressed], 1 + button)
                        self.button_action(device_id, 0, (x, y), 1 + button, pressed)
                self.rfb_buttons = buttons

        GLib.idle_add(process_pointer_event)

    def _process_rfb_KeyEvent(self, proto, packet):
        if not features.input_devices or self.readonly:
            return
        source = self.get_server_source(proto)
        if not source:
            return
        pressed, p1, p2, key = packet[1:5]
        GLib.idle_add(self.process_rfb_key_event, source, pressed, p1, p2, key)

    def process_rfb_key_event(self, source, pressed, p1, p2, key):
        wid = self._get_rfb_desktop_wid()
        keyname = RFB_KEYNAMES.get(key)
        keylog("RFB KeyEvent(%s, %s, %s, %s) keyname=%s, desktop wid=%s", pressed, p1, p2, key, keyname, wid)
        if not keyname:
            if 0 < key < 255:
                keyname = chr(key)
            elif self.X11Keyboard:
                keyname = self.X11Keyboard.keysym_str(key)
        if not keyname:
            keylog.warn("Warning: unknown rfb KeyEvent: %s, %i, %i, %#x", pressed, p1, p2, key)
            return
        modifiers = []
        keyval = 0
        keycode, group = source.keyboard_config.get_keycode(0, keyname, pressed, modifiers, 0, keyname, 0)
        keylog("RFB keycode(%s)=%s, %s", keyname, keycode, group)
        if keycode:
            is_mod = source.keyboard_config.is_modifier(keycode)
            self._handle_key(wid, bool(pressed), keyname, keyval, keycode, modifiers, is_mod, True)

    def _process_rfb_SetEncodings(self, proto, packet):
        encodings = packet[3]
        self._server_sources[proto].set_encodings(encodings)

    def _process_rfb_SetPixelFormat(self, proto, packet):
        pixel_format = packet[4:14]
        self._server_sources[proto].set_pixel_format(pixel_format)

    def _process_rfb_FramebufferUpdateRequest(self, _proto, packet):
        # pressed, _, _, keycode = packet[1:5]
        inc, x, y, w, h = packet[1:6]
        log("RFB: FramebufferUpdateRequest inc=%s, geometry=%s", inc, (x, y, w, h))
        if not inc:
            model = self._get_rfb_desktop_model()
            GLib.idle_add(self.refresh_window_area, model, x, y, w, h)

    def _process_rfb_ClientCutText(self, _proto, packet):
        # l = packet[4]
        text = packet[5]
        log("RFB got clipboard text: %r", text)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

import os
from time import monotonic
from typing import Any, Tuple, List, Dict
from collections.abc import Sequence

from xpra.log import Logger
log = Logger("decoder", "vpx")

from xpra.codecs.constants import VideoSpec, get_subsampling_divs
from xpra.codecs.image import ImageWrapper
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool

from libc.stdint cimport uintptr_t, uint8_t
from libc.string cimport memset, memcpy
from libc.stdlib cimport malloc
from xpra.codecs.vpx.vpx cimport (
    vpx_img_fmt_t, vpx_codec_iface_t,
    vpx_codec_iter_t, vpx_codec_flags_t, vpx_codec_err_t,
    vpx_codec_ctx_t,
    vpx_codec_error, vpx_codec_destroy,
    vpx_codec_version_str, vpx_codec_build_config,
    VPX_IMG_FMT_I420, VPX_IMG_FMT_I444, VPX_IMG_FMT_HIGHBITDEPTH,
    VPX_CS_UNKNOWN, VPX_CS_BT_601, VPX_CS_BT_709,
    VPX_CS_SMPTE_170, VPX_CS_SMPTE_240, VPX_CS_BT_2020,
    VPX_CS_RESERVED, VPX_CS_SRGB,
    vpx_image_t, vpx_color_space_t, vpx_color_range_t,
    VPX_CR_STUDIO_RANGE, VPX_CR_FULL_RANGE,
)
from xpra.buffers.membuf cimport padbuf, MemBuf, buffer_context  # pylint: disable=syntax-error


SAVE_TO_FILE = envbool("XPRA_SAVE_TO_FILE")

VPX_COLOR_SPACES : Dict[int,str] = {
    VPX_CS_UNKNOWN  : "unknown",
    VPX_CS_BT_601   : "BT601",
    VPX_CS_BT_709   : "BT709",
    VPX_CS_SMPTE_170    : "SMPTE170",
    VPX_CS_SMPTE_240    : "SMPTE240",
    VPX_CS_BT_2020  : "BT2020",
    VPX_CS_RESERVED : "reserved",
    VPX_CS_SRGB     : "SRGB",
}

VPX_COLOR_RANGES : Dict[int,str] = {
    VPX_CR_STUDIO_RANGE : "studio",
    VPX_CR_FULL_RANGE   : " full",
}


cdef unsigned int cpus = os.cpu_count()
cdef int VPX_THREADS = envint("XPRA_VPX_THREADS", max(1, cpus-1))


cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)


cdef extern from "vpx/vp8dx.h":
    const vpx_codec_iface_t *vpx_codec_vp8_dx()
    const vpx_codec_iface_t *vpx_codec_vp9_dx()

cdef extern from "vpx/vpx_decoder.h":
    ctypedef struct vpx_codec_enc_cfg_t:
        unsigned int rc_target_bitrate
        unsigned int g_lag_in_frames
        unsigned int rc_dropframe_thresh
        unsigned int rc_resize_allowed
        unsigned int g_w
        unsigned int g_h
        unsigned int g_error_resilient
    ctypedef struct vpx_codec_dec_cfg_t:
        unsigned int threads
        unsigned int w
        unsigned int h
    cdef int VPX_CODEC_OK
    cdef int VPX_DECODER_ABI_VERSION

    vpx_codec_err_t vpx_codec_dec_init_ver(vpx_codec_ctx_t *ctx, vpx_codec_iface_t *iface,
                                            vpx_codec_dec_cfg_t *cfg, vpx_codec_flags_t flags, int ver)

    vpx_codec_err_t vpx_codec_decode(vpx_codec_ctx_t *ctx, const uint8_t *data,
                                     unsigned int data_sz, void *user_priv, long deadline) nogil

    vpx_image_t *vpx_codec_get_frame(vpx_codec_ctx_t *ctx, vpx_codec_iter_t *citer) nogil


#https://groups.google.com/a/webmproject.org/forum/?fromgroups#!msg/webm-discuss/f5Rmi-Cu63k/IXIzwVoXt_wJ
#"RGB is not supported.  You need to convert your source to YUV, and then compress that."
COLORSPACES : Dict[str, Sequence[str]] = {
    "vp8"   : ("YUV420P", ),
    "vp9"   : ("YUV420P", "YUV444P"),
}
CODECS = tuple(COLORSPACES.keys())


def get_abi_version() -> int:
    return VPX_DECODER_ABI_VERSION


def get_version() -> Sequence[int]:
    b = vpx_codec_version_str()
    vstr = b.decode("latin1").lstrip("v")
    log("vpx_codec_version_str()=%s", vstr)
    vparts : List[int] = []
    try:
        for x in vstr.split("."):
            vparts.append(int(x))
    except Exception:
        pass
    return tuple(vparts)


def get_type() -> str:
    return "vpx"


def get_encodings() -> Sequence[str]:
    return CODECS


def get_min_size(encoding:str) -> Tuple[int, int]:
    return 16, 16


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    for encoding, in_css in COLORSPACES.items():
        for colorspace in in_css:
            specs.append(VideoSpec(
                    encoding=encoding, input_colorspace=colorspace, output_colorspaces=(colorspace, ),
                    has_lossless_mode=encoding == "vp9" and colorspace == "YUV444P",
                    codec_class=Decoder, codec_type=get_type(),
                    quality=50, speed=50,
                    size_efficiency=60,
                    setup_cost=20,
                    max_w=8192,
                    max_h=4096,
                )
            )
    return specs


def get_info() -> Dict[str, Any]:
    global CODECS
    info = {
        "version"       : get_version(),
        "encodings"     : CODECS,
        "abi_version"   : get_abi_version(),
        "build_config"  : vpx_codec_build_config(),
    }
    for k,v in COLORSPACES.items():
        info["%s.colorspaces" % k] = v
    return info


cdef inline const vpx_codec_iface_t  *make_codec_dx(encoding):
    if encoding=="vp8":
        return vpx_codec_vp8_dx()
    if encoding=="vp9":
        return vpx_codec_vp9_dx()
    raise ValueError(f"unsupported encoding: {encoding!r}")


cdef inline vpx_img_fmt_t get_vpx_colorspace(colorspace):
    if colorspace == "YUV444P":
        return VPX_IMG_FMT_I444
    return VPX_IMG_FMT_I420


cdef class Decoder:

    cdef vpx_codec_ctx_t *context
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned int max_threads
    cdef unsigned long frames
    cdef vpx_img_fmt_t pixfmt
    cdef object dst_format
    cdef object encoding
    cdef object file

    cdef object __weakref__

    def init_context(self, encoding: str, width: int, height: int, colorspace: str, options: typedict) -> None:
        log("vpx decoder init_context%s", (encoding, width, height, colorspace))
        assert encoding in CODECS, f"invalid encoding {encoding!r}"
        assert colorspace in COLORSPACES[encoding], f"invalid colorspace {colorspace!r} for encoding {encoding!r}"
        cdef int flags = 0
        cdef const vpx_codec_iface_t *codec_iface = make_codec_dx(encoding)
        self.encoding = encoding
        self.dst_format = colorspace
        self.pixfmt = get_vpx_colorspace(self.dst_format)
        self.width = width
        self.height = height
        try:
            #no point having too many threads if the height is small, also avoids a warning:
            self.max_threads = max(0, min(int(VPX_THREADS), roundup(height, 32)//32*2))
        except:
            self.max_threads = 1
        self.context = <vpx_codec_ctx_t *> malloc(sizeof(vpx_codec_ctx_t))
        assert self.context!=NULL
        memset(self.context, 0, sizeof(vpx_codec_ctx_t))
        cdef vpx_codec_dec_cfg_t dec_cfg
        dec_cfg.w = width
        dec_cfg.h = height
        dec_cfg.threads = self.max_threads
        if vpx_codec_dec_init_ver(self.context, codec_iface, &dec_cfg,
                              flags, VPX_DECODER_ABI_VERSION)!=VPX_CODEC_OK:
            raise RuntimeError("failed to instantiate %s decoder with ABI version %s: %s" % (encoding, VPX_DECODER_ABI_VERSION, self.codec_error_str()))
        if SAVE_TO_FILE:
            filename = f"./{monotonic()}.{self.encoding}"
            self.file = open(filename, "wb")
            log.info("saving %s stream to %s", self.encoding, filename)
        log("vpx_codec_dec_init_ver for %s succeeded with ABI version %s", encoding, VPX_DECODER_ABI_VERSION)

    def __repr__(self):
        return "vpx.Decoder(%s)" % self.encoding

    def get_info(self) -> Dict[str, Any]:
        return {
            "type"      : self.get_type(),
            "width"     : self.get_width(),
            "height"    : self.get_height(),
            "encoding"  : self.encoding,
            "frames"    : int(self.frames),
            "colorspace": self.get_colorspace(),
            "max_threads" : self.max_threads,
        }

    def get_colorspace(self) -> str:
        return self.dst_format

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def is_closed(self) -> bool:
        return self.context==NULL

    def get_encoding(self) -> str:
        return self.encoding

    def get_type(self) -> str:
        return  "vpx"

    def __dealloc__(self):
        self.clean()

    def clean(self) -> None:
        if self.context!=NULL:
            vpx_codec_destroy(self.context)
            self.context = NULL
        self.width = 0
        self.height = 0
        self.max_threads = 0
        self.pixfmt = 0
        self.dst_format = ""
        self.encoding = ""
        f = self.file
        if f:
            self.file = None
            f.close()

    def decompress_image(self, data: bytes, options: typedict) -> ImageWrapper:
        cdef vpx_codec_iter_t citer = NULL
        cdef MemBuf output_buf
        cdef void *output
        cdef Py_ssize_t plane_len
        cdef uint8_t dy
        cdef unsigned int height
        cdef int stride
        assert self.context!=NULL

        cdef double start = monotonic()
        cdef vpx_codec_err_t ret = -1
        cdef uint8_t* src
        cdef Py_ssize_t src_len
        with buffer_context(data) as bc:
            src = <uint8_t*> (<uintptr_t> int(bc))
            src_len = len(bc)
            with nogil:
                ret = vpx_codec_decode(self.context, src, src_len, NULL, 0)
        if ret!=VPX_CODEC_OK:
            log.error("Error: vpx_codec_decode: %s", self.codec_error_str())
            return None
        cdef vpx_image_t *img
        with nogil:
            img = vpx_codec_get_frame(self.context, &citer)
        if img==NULL:
            log.error("Error: vpx_codec_get_frame: %s", self.codec_error_str())
            return None
        if img.fmt != self.pixfmt:
            expected = self.dst_format
            if img.fmt == VPX_IMG_FMT_I444:
                self.dst_format = "YUV444P"
            elif img.fmt == VPX_IMG_FMT_I420:
                self.dst_format = "YUV420P"
            else:
                raise RuntimeError("unexpected image pixel format %s" % img.fmt)
            log.warn(f"Warning: expected {expected} but got {self.dst_format}")
        strides = []
        pixels = []
        divs = get_subsampling_divs(self.get_colorspace())
        cdef int i
        for i in range(3):
            _, dy = divs[i]
            if dy==1:
                height = self.height
            elif dy==2:
                height = (self.height+1)>>1
            else:
                raise ValueError(f"invalid height divisor {dy} for {self.get_colorspace()}")
            stride = img.stride[i]
            strides.append(stride)

            plane_len = height * stride
            #add one extra line of padding:
            output_buf = padbuf(plane_len, stride, 0)
            output = <void *>output_buf.get_mem()
            memcpy(output, <void *>img.planes[i], plane_len)
            memset(<void *>((<char *>output)+plane_len), 0, stride)

            pixels.append(memoryview(output_buf))
        self.frames += 1
        cdef double elapsed = 1000*(monotonic()-start)
        log("%s frame %4i decoded in %3ims, colorspace=%s, format=%s",
            self.encoding, self.frames, elapsed, VPX_COLOR_SPACES.get(img.cs, img.cs), self.dst_format)
        image = ImageWrapper(0, 0, self.width, self.height, pixels, self.get_colorspace(), 24, strides, 1, ImageWrapper.PLANAR_3)
        image.set_full_range(options.boolget("full-range", True))
        return image

    def codec_error_str(self) -> str:
        return vpx_codec_error(self.context).decode("latin1")


def selftest(full=False) -> None:
    global CODECS
    from xpra.codecs.checks import testdecoder
    from xpra.codecs.vpx import decoder
    CODECS = testdecoder(decoder, full)



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# xxhash wrapper

#cython: wraparound=False
from libc.stdint cimport uint64_t

cdef extern from "xxhash.h":
    ctypedef uint64_t XXH64_hash_t
    XXH64_hash_t XXH3_64bits(const void* data, size_t length) nogil

cdef uint64_t xxh3(const void* data, size_t length) noexcept nogil:
    return XXH3_64bits(data, length)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any

from xpra.log import Logger

log = Logger("audio")


def has_pa() -> bool:
    return False


def set_source_mute(device, mute=False) -> None:
    """ none implementation """


def set_sink_mute(device, mute=False) -> None:
    """ none implementation """


def get_default_sink() -> str:
    return ""


def get_pulse_server() -> str:
    return ""


def get_pulse_id() -> str:
    return ""


def get_pulse_cookie_hash() -> str:
    return ""


def get_pactl_server() -> str:
    return ""


def get_pa_device_options(*_args) -> dict[str, Any]:
    return {}


def get_info() -> dict[str, Any]:
    return {
        "pulseaudio.wrapper": "none",
        "pulseaudio.found": has_pa(),
    }


def main():
    if "-v" in sys.argv:
        log.enable_debug()
    i = get_info()
    for k in sorted(i):
        log.info("%s : %s", k.ljust(64), i[k])


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from collections.abc import Sequence

from xpra.net.digest import get_salt, choose_digest
from xpra.auth.sys_auth_base import SysAuthenticator
from xpra.log import Logger

log = Logger("auth")


def stat_filetime(full_path) -> float:
    try:
        v = os.stat(full_path).st_mtime
        log("mtime(%s)=%s", full_path, v)
        return v
    except Exception as e:
        log.error(f"Error accessing time of file {full_path!r}")
        log.estr(e)
        return 0


class FileAuthenticatorBase(SysAuthenticator):
    def __init__(self, **kwargs):
        password_file = kwargs.pop("filename", "")
        log("FileAuthenticatorBase password_file=%s", password_file)
        if not password_file:
            log.warn("Warning: %r authentication module is missing the 'filename' option", self)
            log.warn(" all authentication attempts will fail")
        elif not os.path.isabs(password_file):
            exec_cwd = kwargs.get("exec_cwd", os.getcwd())
            password_file = os.path.join(exec_cwd, password_file)
        log("FileAuthenticatorBase filename=%s", password_file)
        super().__init__(**kwargs)
        self.salt: bytes | bool | None = None
        self.digest: str = ""
        self.challenge_sent = False
        self.password_filename = password_file
        self.password_filedata = None
        self.password_filetime = 0.0
        self.authenticate_check = self.authenticate_hmac

    def requires_challenge(self) -> bool:
        return True

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if self.salt is not None:
            log.error("challenge already sent!")
            if self.salt is not False:
                self.salt = False
            return b"", ""
        self.salt = get_salt()
        self.digest = choose_digest(digests)
        self.challenge_sent = True
        return self.salt, self.digest

    def get_password(self) -> str:
        file_data = self.load_password_file()
        if not file_data:
            return ""
        return file_data

    def parse_filedata(self, data: str):  # pragma: no cover
        raise NotImplementedError()

    def load_password_file(self):
        if not self.password_filename:
            return ""
        full_path = os.path.abspath(self.password_filename)
        if not os.path.exists(self.password_filename):
            log.error("Error: password file '%s' is missing", full_path)
            self.password_filedata = ""
        else:
            ptime = self.stat_password_filetime()
            if not self.password_filedata or ptime != self.password_filetime:
                self.password_filetime = 0
                self.password_filedata = ""
                try:
                    with open(self.password_filename, encoding="utf8") as f:
                        data = f.read()
                    log(f"loaded {len(data)} characters from {self.password_filename!r}")
                    self.password_filedata = self.parse_filedata(data)
                    self.password_filetime = ptime
                except Exception as e:
                    log.error("Error reading password data from '%s':", self.password_filename, exc_info=True)
                    log.estr(e)
                    self.password_filedata = self.parse_filedata("")
        return self.password_filedata

    def stat_password_filetime(self) -> float:
        full_path = os.path.abspath(self.password_filename or "")
        return stat_filetime(full_path)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import OSX, POSIX, gi_import
from xpra.util.env import first_time, IgnoreWarningsContext
from xpra.util.system import is_X11
from xpra.log import Logger

Gdk = gi_import("Gdk")


def get_default_root_window() -> Gdk.Window | None:
    screen = Gdk.Screen.get_default()
    if screen is None:
        return None
    return screen.get_root_window()


def get_root_size(default: None | tuple[int, int] = (1920, 1024)) -> tuple[int, int] | None:
    if OSX:
        # the easy way:
        root = get_default_root_window()
        if not root:
            return default
        w, h = root.get_geometry()[2:4]
    else:
        # GTK3 on win32 triggers this warning:
        # "GetClientRect failed: Invalid window handle."
        # if we try to use the root window,
        # and on Linux with Wayland, we get bogus values...
        screen = Gdk.Screen.get_default()
        if screen is None:
            return default
        with IgnoreWarningsContext():
            w = screen.get_width()
            h = screen.get_height()
    if w <= 0 or h <= 0 or w > 32768 or h > 32768:
        if first_time("Gtk root window dimensions"):
            log = Logger("gtk", "screen")
            log.warn(f"Warning: Gdk returned invalid root window dimensions: {w}x{h}")
            log.warn(" no access to the display?")
            log.warn(f" using {default} instead")
        return default
    return w, h


GRAB_STATUS_STRING = {
    Gdk.GrabStatus.SUCCESS: "SUCCESS",
    Gdk.GrabStatus.ALREADY_GRABBED: "ALREADY_GRABBED",
    Gdk.GrabStatus.INVALID_TIME: "INVALID_TIME",
    Gdk.GrabStatus.NOT_VIEWABLE: "NOT_VIEWABLE",
    Gdk.GrabStatus.FROZEN: "FROZEN",
}

dsinit: bool = False


def init_display_source() -> None:
    """
    On X11, we want to be able to access the bindings,
    so we need to get the X11 display from GDK.
    """
    global dsinit
    dsinit = True
    x11 = is_X11()
    log = Logger("gtk", "screen")
    log(f"init_display_source() {x11=}")
    if x11:
        try:
            from xpra.x11.gtk.display_source import init_gdk_display_source
            init_gdk_display_source()
        except ImportError:  # pragma: no cover
            log("init_gdk_display_source()", exc_info=True)
            log.warn("Warning: the Gtk-3.0 X11 bindings are missing")
            log.warn(" some features may be degraded or unavailable")
            log.warn(" ie: keyboard mapping, focus, etc")


def ds_inited() -> bool:
    return dsinit


def main():
    from xpra.platform import program_context
    from xpra.util.str_fn import print_nested_dict
    from xpra.log import enable_color
    with program_context("GTK-Version-Info", "GTK Version Info"):
        enable_color()
        from xpra.gtk.versions import get_gtk_version_info
        print("%s" % get_gtk_version_info())
        if POSIX and not OSX:
            from xpra.x11.bindings.posix_display_source import init_posix_display_source
            init_posix_display_source()
        import warnings
        warnings.simplefilter("ignore")
        from xpra.gtk.info import get_display_info, get_screen_sizes
        print(get_screen_sizes()[0])
        print_nested_dict(get_display_info())


if __name__ == "__main__":
    main()
